from typing import TypeVar

from amino import do, Do, Map, List, Maybe
from amino.lenses.lens import lens

from ribosome.util.menu.auto.data import AutoS
from ribosome.nvim.io.state import NS
from ribosome.util.menu.data import MenuPrompt, MenuUpdateCursor, MenuLine
from ribosome.util.menu.prompt.data import PromptPassthrough

A = TypeVar('A')


# TODO fork by prompt state. if in passthrough, esc should quit
@do(AutoS)
def menu_cmd_esc() -> Do:
    yield NS.pure(MenuPrompt(PromptPassthrough()))


@do(AutoS)
def menu_cmd_up() -> Do:
    yield NS.modify(lens.data.cursor.modify(lambda a: max(a - 1, 0)))
    return MenuUpdateCursor()


@do(AutoS)
def menu_cmd_down() -> Do:
    count = yield NS.inspect(lambda a: a.data.content.filtered.length)
    yield NS.modify(lens.data.cursor.modify(lambda a: min(a + 1, count)))
    return MenuUpdateCursor()


def toggle_selected(lines: List[MenuLine[A]], index: int) -> Maybe[List[MenuLine[A]]]:
    return lines.modify_at(index, lambda b: b.mod.selected(lambda c: not c))


visibility_error = 'broken visibility map for menu'


@do(AutoS)
def menu_cmd_select_cursor() -> Do:
    cursor = yield NS.inspect(lambda a: a.cursor)
    content = yield NS.inspect(lambda a: a.data.content)
    index = yield NS.from_maybe(content.visible.lift(cursor), lambda: visibility_error)
    toggled = yield NS.from_maybe(toggle_selected(content.lines, index), lambda: visibility_error)
    yield NS.modify(lens.data.content.lines.set(toggled))
    return MenuUpdateCursor()


@do(AutoS)
def menu_cmd_select_all() -> Do:
    yield NS.modify(lens.data.content.lines.modify(lambda a: a.map(lambda b: b.mod.selected(lambda c: not c))))
    return MenuUpdateCursor()


builtin_mappings = Map({
    '<esc>': menu_cmd_esc,
    'k': menu_cmd_up,
    'j': menu_cmd_down,
    '<space>': menu_cmd_select_cursor,
    '*': menu_cmd_select_all,
})

__all__ = ('builtin_mappings',)



from typing import Any

from amino import Dat, List, Nil

from ribosome.rpc.data.rpc_type import NonblockingRpc, RpcType, BlockingRpc


class Rpc(Dat['Rpc']):

    @staticmethod
    def nonblocking(method: str, args: List[Any]) -> 'Rpc':
        return Rpc(method, args, NonblockingRpc())

    def __init__(self, method: str, args: List[Any], tpe: RpcType) -> None:
        self.method = method
        self.args = args
        self.tpe = tpe

    @property
    def sync(self) -> bool:
        return isinstance(self.tpe, BlockingRpc)


class ActiveRpc(Dat['ActiveRpc']):

    def __init__(self, rpc: Rpc, id: int) -> None:
        self.rpc = rpc
        self.id = id


class RpcArgs(Dat['RpcArgs']):

    @staticmethod
    def cons(args: List[Any]=Nil, bang: bool=False) -> 'RpcArgs':
        return RpcArgs(args, bang)

    def __init__(self, args: List[Any], bang: bool) -> None:
        self.args = args
        self.bang = bang

    @property
    def string(self) -> str:
        return self.args.join_comma


__all__ = ('Rpc', 'ActiveRpc', 'RpcArgs',)



from typing import TypeVar, Callable, Any

from kallikrein import kf
from kallikrein.expectable import Expectable

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIO

A = TypeVar('A')


def kn(vim: NvimApi, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
    return kf(lambda: f(*a, **kw).result(vim))


def kns(vim: NvimApi, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
    return kf(lambda: f(*a, **kw).run_s(vim))


__all__ = ('kn', 'kns')



from typing import Callable

from amino import do, curried, Do, __, _, Either
from amino.lenses.lens import lens
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.program import Program
from ribosome.config.component import Components
from ribosome.nvim.api.command import nvim_command
from ribosome.data.mapping import Mapping, MapMode

log = module_log()


def mapping_handler(mapping: Mapping) -> Callable[[Components], Either[str, Program]]:
    def mapping_handler(components: Components) -> Either[str, Program]:
        return components.all.find_map(__.mappings.lift(mapping)).to_either(f'no handler for {mapping}')
    return mapping_handler


def mapping_cmd(plugin: str, mapping: Mapping, mode: MapMode) -> NvimIO[None]:
    buf = '<buffer>' if mapping.buffer else ''
    keys = mapping.keys.replace('<', '<lt>')
    rhs = f''':call {plugin}Map('{mapping.ident}', '{keys}')<cr>'''
    return nvim_command(
        f'{mode.mnemonic}map',
        buf,
        '<silent>',
        mapping.keys,
        rhs,
    )


@do(NS[PluginState, None])
def activate_mapping(mapping: Mapping) -> Do:
    handler = yield NS.inspect_either(mapping_handler(mapping)).zoom(lens.components)
    yield NS.modify(__.append.active_mappings((mapping.ident, handler)))
    plugin = yield NS.inspect(_.camelcase_name)
    yield NS.lift(mapping.modes.traverse(curried(mapping_cmd)(plugin, mapping), NvimIO))


__all__ = ('activate_mapping',)



from typing import Any, TypeVar, Callable

from ribosome.nvim.io.compute import NvimIO, NRParams

from amino import List, do, Do, Either, Lists, Right, Left
from amino.logging import module_log
from ribosome.nvim.api.util import cons_split_lines, cons_checked_e
from ribosome.nvim.io.api import N

log = module_log()
A = TypeVar('A')


def nvim_command(cmd: str, *args: Any, params: NRParams=NRParams.cons(verbose=False, sync=False)) -> NvimIO[None]:
    arg_string = ' '.join(map(str, args))
    arg_suffix = '' if len(args) == 0 else f' {arg_string}'
    silent = '' if params.verbose else 'silent! '
    cmdline = f'{silent}{cmd}{arg_suffix}'
    log.debug1(lambda: f'nvim command `{cmdline}`')
    return N.write('nvim_command', cmdline, params=params)


def nvim_sync_command(cmd: str, *args: Any, verbose: bool=False) -> NvimIO[None]:
    return nvim_command(cmd, *args, verbose=verbose, sync=True)


def nvim_command_output(cmd: str, *args: Any) -> NvimIO[None]:
    cmdline = ' '.join(map(str, (cmd,) + args))
    return N.read_cons_strict('nvim_command_output', cons_split_lines, cmdline)


def doautocmd(name: str, pattern: str='', params: NRParams=NRParams.cons(verbose=False, sync=False)) -> NvimIO[None]:
    return nvim_command(f'doautocmd', '<nomodeline>', name, pattern, params=params)


def runtime(path: str, params: NRParams=NRParams.cons(verbose=True, sync=False)) -> NvimIO[None]:
    return nvim_command('runtime!', f'{path}.vim', params=params)


def defined_commands() -> NvimIO[List[str]]:
    return nvim_command_output('command')


@do(NvimIO[str])
def defined_commands_str() -> Do:
    lines = yield defined_commands()
    return lines.join_lines


@do(Either[str, None])
def atomic_error(cmdlines: List[str], error_raw: Any) -> Do:
    error = yield (
        Right(error_raw)
        if isinstance(error_raw, list) else
        Left(f'invalid error structure for atomic call: {error_raw}')
    )
    index, tpe, message = yield (
        Lists.wrap(error)
        .lift_all(0, 1, 2)
        .to_either(f'too few elements in error structure for atomic call: {error}')
    )
    offender = yield cmdlines.lift(index).to_either(f'invalid index in atomic call error: {error}')
    yield Left(f'error of type `{tpe}` in atomic call `{offender}`: {message}')


def cons_atomic_result(cmdlines: List[str]) -> Callable[[list], Either[str, Any]]:
    @do(Either[str, Any])
    def cons(raw: list) -> Do:
        result = Lists.wrap(raw)
        results, error = yield result.lift_all(0, 1).to_either_f(lambda: f'too few elements in atomic result: {result}')
        yield (
            Right(results)
            if error is None
            else atomic_error(cmdlines, error)
        )
    return cons


def nvim_atomic_commands(cmdlines: List[str]) -> NvimIO[List[Any]]:
    cmds = cmdlines.map(lambda a: ['nvim_command', [a]])
    return N.read_cons_strict('nvim_call_atomic', cons_checked_e(list, cons_atomic_result(cmdlines)), cmds)


__all__ = ('nvim_command', 'nvim_command_output', 'doautocmd', 'runtime', 'nvim_sync_command', 'defined_commands',
           'defined_commands_str', 'nvim_atomic_commands',)



from concurrent.futures import Future
from typing import Any, Callable, TypeVar
from threading import Lock

from amino import do, Do, IO, Map, Dat
from amino.logging import module_log

from ribosome.rpc.error import RpcReadError
from ribosome.rpc.data.rpc import ActiveRpc

A = TypeVar('A')
log = module_log()


PendingRpc = Map[int, Future]


class Requests(Dat['Requests']):

    @staticmethod
    def cons(current_id: int=0, to_vim: PendingRpc=Map(), from_vim: PendingRpc=Map()) -> 'Requests':
        return Requests(current_id, to_vim, from_vim)

    def __init__(self, current_id: int, to_vim: PendingRpc, from_vim: PendingRpc) -> None:
        self.current_id = current_id
        self.to_vim = to_vim
        self.from_vim = from_vim


OnMessage = Callable[[bytes], IO[None]]
OnError = Callable[[RpcReadError], IO[None]]


class RpcConcurrency(Dat['RpcConcurrency']):

    @staticmethod
    def cons(
            requests: Requests=None,
            lock: Lock=None,
    ) -> 'RpcConcurrency':
        return RpcConcurrency(
            requests or Requests.cons(),
            lock or Lock(),
        )

    def exclusive(self, f: Callable[..., IO[A]], *a: Any, **kw: Any) -> IO[A]:
        def wrap() -> IO[A]:
            with self.lock:
                return IO.from_either(f(*a, **kw).attempt)
        return IO.suspend(wrap)

    def __init__(self, requests: Requests, lock: Lock) -> None:
        self.requests = requests
        self.lock = lock


def exclusive_unregister_rpc(rc: RpcConcurrency, requests: PendingRpc, rpc: ActiveRpc) -> IO[Future]:
    return IO.delay(requests.pop, rpc.id)


def unregister_rpc(rc: RpcConcurrency, requests: PendingRpc, rpc: ActiveRpc) -> IO[Future]:
    log.debug1(f'unregistering {rpc}')
    return (
        IO.failed(f'invalid request id from vim after execution: {rpc}. active requests: {requests}')
        if rpc.id not in requests else
        rc.exclusive(exclusive_unregister_rpc, rc, requests, rpc)
    )


@do(IO[Future])
def exclusive_register_rpc(rc: RpcConcurrency, requests: PendingRpc, rpc: ActiveRpc) -> Do:
    f: Future = Future()
    yield IO.delay(requests.update, {rpc.id: f})
    return f


@do(IO[Future])
def register_rpc(rc: RpcConcurrency, requests: PendingRpc, rpc: ActiveRpc) -> Do:
    log.debug1(f'registering {rpc}')
    yield (
        IO.failed(f'duplicate id in request from vim: {rpc}')
        if rpc.id in requests else
        rc.exclusive(exclusive_register_rpc, rc, requests, rpc)
    )

__all__ = ('Requests', 'OnMessage', 'OnError', 'RpcConcurrency', 'unregister_rpc', 'register_rpc',)



from typing import Callable

from amino import List, do, Do, Dat, Nil, Just, Nothing, Maybe
from amino.case import Case
from amino.util.string import camelcase
from amino.logging import module_log

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.rpc import channel_id
from ribosome.nvim.api.command import nvim_atomic_commands
from ribosome.rpc.api import RpcOptions, RpcProgram
from ribosome.rpc.data.rpc_method import RpcMethod, CommandMethod, FunctionMethod, AutocmdMethod
from ribosome.rpc.data.prefix_style import PrefixStyle, Plain, Full, Short
from ribosome.rpc.data.nargs import Nargs, NargsStar

log = module_log()


# FIXME do multiple handlers work?
# @property
# def distinct_specs(self) -> List[RpcProgram]:
#     return self.specs.distinct_by(_.rpc_method)


class RpcConfig(Dat['RpcConfig']):

    def __init__(self, channel: int, name: str, prefix: str) -> None:
        self.channel = channel
        self.name = name
        self.prefix = prefix


class RpcDef(Dat['RpcDef']):

    def __init__(
            self,
            config: RpcConfig,
            options: RpcOptions,
            prog_name: str,
            rpc_name: str,
            trigger_name: str,
    ) -> None:
        self.config = config
        self.options = options
        self.prog_name = prog_name
        self.rpc_name = rpc_name
        self.trigger_name = trigger_name


class ActiveRpcTrigger(Dat['ActiveRpcTrigger']):

    def __init__(self, name: str, prog: RpcProgram, method: RpcMethod, channel: int, definition: List[str]) -> None:
        self.name = name
        self.prog = prog
        self.method = method
        self.channel = channel
        self.definition = definition


def quote(a: str) -> str:
    return f'\'{a}\''


class undef_command(Case[RpcMethod, Maybe[str]], alg=RpcMethod):

    def function(self, method: FunctionMethod) -> Maybe[str]:
        return Just('delfunction')

    def command(self, method: CommandMethod) -> Maybe[str]:
        return Just('delcommand')

    def autocmd(self, method: AutocmdMethod) -> Maybe[str]:
        return Nothing


class DefinitionTokens(Dat['DefinitionTokens']):

    def __init__(
            self,
            rpc_function: str,
            rpc_args: List[str],
            def_cmd: str,
            rpc_opts_pre: List[str],
            rpc_opts_post: List[str],
            method_prefix: str,
    ) -> None:
        self.rpc_function = rpc_function
        self.rpc_args = rpc_args
        self.def_cmd = def_cmd
        self.rpc_opts_pre = rpc_opts_pre
        self.rpc_opts_post = rpc_opts_post
        self.method_prefix = method_prefix



def command_tokens(bang: bool, nargs: Nargs) -> DefinitionTokens:
    bang_opt = List('-bang') if bang else Nil
    bang_arg = List("<q-bang> == '!'") if bang else Nil
    nargs_arg = List('[<f-args>]')
    opts = List(f'-nargs={nargs.for_vim}') + bang_opt
    args = bang_arg + nargs_arg
    return DefinitionTokens('rpcnotify', args, 'command!', opts, Nil, '')


function_tokens = DefinitionTokens('rpcrequest', List('a:000'), 'function!', Nil, Nil, '')


def autocmd_tokens(pattern: str, sync: bool) -> DefinitionTokens:
    func = 'rpcrequest' if sync else 'rpcnotify'
    return DefinitionTokens(func, Nil, 'autocmd', Nil, List('*'), '')


class rpc_prefix(Case[PrefixStyle, str], alg=PrefixStyle):

    def __init__(self, name: str, prefix: str, method: str) -> None:
        self.name = name
        self.prefix = prefix
        self.method = method

    def plain(self, style: Plain) -> str:
        return self.method

    def full(self, style: Full) -> str:
        return f'{self.name}_{self.method}'

    def short(self, style: Short) -> str:
        return f'{self.prefix}_{self.method}'


def rpc_trigger_name(config: RpcConfig, options: RpcOptions, rpc_name: str) -> str:
    prefixed = rpc_prefix(config.name, config.prefix, rpc_name)(options.prefix)
    return camelcase(prefixed)


def define_trigger_tokens(
        rhs_f: Callable[[str], str],
        rpc_def: RpcDef,
        tokens: DefinitionTokens,
) -> List[str]:
    method = f'{tokens.method_prefix}{rpc_def.rpc_name}'
    args = tokens.rpc_args.cons(quote(method)).join_comma
    rpc_call = f'{tokens.rpc_function}({rpc_def.config.channel}, {args})'
    rhs = rhs_f(rpc_call)
    return (
        List(tokens.def_cmd) +
        tokens.rpc_opts_pre +
        List(rpc_def.trigger_name) +
        tokens.rpc_opts_post +
        List(rhs)
    )


def trigger_definition(
        rhs: Callable[[str], str],
        rpc_def: RpcDef,
        tokens: DefinitionTokens,
) -> str:
    return define_trigger_tokens(rhs, rpc_def, tokens).join_tokens


def command_rhs(a: str) -> str:
    return f'call {a}'


def function_rhs(a: str) -> str:
    return f'(...)\nreturn {a}\nendfunction'


class method_definition(Case[RpcMethod, List[str]], alg=RpcMethod):

    def __init__(self, prog: RpcProgram, rpc_def: RpcDef) -> None:
        self.prog = prog
        self.rpc_def = rpc_def

    def command(self, method: CommandMethod) -> List[str]:
        nargs = (
            NargsStar()
            if self.prog.options.json else
            self.prog.program.params_spec.nargs
        )
        tokens = command_tokens(method.bang, nargs)
        return List(trigger_definition(command_rhs, self.rpc_def, tokens))

    def function(self, method: FunctionMethod) -> List[str]:
        return List(trigger_definition(function_rhs, self.rpc_def, function_tokens))

    def autocmd(self, method: AutocmdMethod) -> List[str]:
        tokens = autocmd_tokens(method.pattern, method.sync)
        return List(
            f'augroup {self.rpc_def.config.name}',
            trigger_definition(command_rhs, self.rpc_def, tokens),
            f'augroup end'
        )


def prog_triggers(rpc_config: RpcConfig) -> Callable[[RpcProgram], ActiveRpcTrigger]:
    def traverse(prog: RpcProgram) -> ActiveRpcTrigger:
        def define(method: RpcMethod) -> List[str]:
            log.debug1(lambda: f'defining {prog} for {method} on channel {rpc_config.channel}')
            trigger_name = rpc_trigger_name(rpc_config, prog.options, prog.rpc_name)
            rpc_def = RpcDef(rpc_config, prog.options, prog.program.name, prog.rpc_name, trigger_name)
            return ActiveRpcTrigger(trigger_name, prog, method, rpc_config.channel,
                                    method_definition(prog, rpc_def)(method))
        return prog.options.methods.map(define)
    return traverse


def rpc_triggers(progs: List[RpcProgram], name: str, prefix: str, channel: int) -> List[ActiveRpcTrigger]:
    return progs.flat_map(prog_triggers(RpcConfig(channel, name, prefix)))


@do(NvimIO[List[ActiveRpcTrigger]])
def define_rpc(progs: List[RpcProgram], name: str, prefix: str) -> Do:
    channel = yield channel_id()
    triggers = rpc_triggers(progs, name, prefix, channel)
    definitions = triggers.flat_map(lambda a: a.definition)
    yield nvim_atomic_commands(definitions)
    return triggers

__all__ = ('define_rpc',)



from typing import Generic, TypeVar

from amino import ADT

from ribosome.nvim.io.state import NS

A = TypeVar('A')
D = TypeVar('D')


class Compilation(Generic[D, A], ADT['Compilation']):
    pass


class CompilationSuccess(Generic[D, A], Compilation[D, A]):

    def __init__(self, prog: NS[D, A]) -> None:
        self.prog = prog


class CompilationFailure(Generic[D, A], Compilation[D, A]):

    def __init__(self, error: str) -> None:
        self.error = error


__all__ = ('Compilation', 'CompilationSuccess', 'CompilationFailure')



from typing import Any, Type, TypeVar, Callable

from ribosome.nvim.io.compute import NvimIO, NRParams

from amino import Either, List
from ribosome.nvim.api.command import nvim_command
from ribosome.nvim.request import nvim_request
from ribosome.nvim.io.api import N
from ribosome.nvim.api.util import cons_json, cons_json_tpe

A = TypeVar('A')


def nvim_call_function(fun: str, *args: Any, params: NRParams=NRParams.cons(sync=True)) -> NvimIO[Any]:
    return nvim_request('nvim_call_function', fun, args, params=params)


def nvim_call_tpe(tpe: Type[A], fun: str, *args: Any) -> NvimIO[A]:
    return N.read_tpe('nvim_call_function', tpe, fun, args)


def nvim_call_cons(
        cons: Callable[[Any], NvimIO[Either[str, A]]],
        fun: str,
        *args: Any,
        params: NRParams=NRParams.cons(decode=True),
) -> NvimIO[A]:
    return N.read_cons('nvim_call_function', cons, fun, args, params=params)


def nvim_call_cons_strict(cons: Callable[[Any], Either[str, A]], fun: str, *args: Any) -> NvimIO[A]:
    return N.read_cons_strict('nvim_call_function', cons, fun, args)


# TODO check result; add `nvim_call_json_list`
def nvim_call_json(fun: str, *args: Any) -> NvimIO[A]:
    return nvim_call_cons_strict(cons_json, fun, *args)


def nvim_call_json_tpe(tpe: Type[A], fun: str, *args: Any) -> NvimIO[A]:
    return nvim_call_cons_strict(cons_json_tpe(tpe), fun, *args)


def define_function(name: str, params: List[str], body: str) -> NvimIO[None]:
    return nvim_command(f'function!', f'{name}({params.join_comma})\n{body}\nendfunction')


__all__ = ('nvim_call_function', 'nvim_call_tpe', 'nvim_call_cons_strict', 'define_function', 'nvim_call_json',
           'nvim_call_json_tpe', 'nvim_call_cons',)



from typing import Any

from ribosome.nvim.api.util import nvimio_repeat_timeout
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.rpc import plugin_name
from ribosome.nvim.api.function import nvim_call_json

from amino import do, List, Do
from amino.util.string import camelcase


@do(NvimIO[List[str]])
def program_log() -> Do:
    name = yield plugin_name()
    yield nvim_call_json(f'{camelcase(name)}ProgramLog')


def seen_program(name: str, timeout: float=1., interval=.25) -> NvimIO[None]:
    return nvimio_repeat_timeout(
        program_log,
        lambda a: a.contains(name),
        f'program `{name}` wasn\'t executed',
        timeout=timeout,
        interval=interval,
    )


@do(NvimIO[Any])
def plugin_state() -> Do:
    name = yield plugin_name()
    yield nvim_call_json(f'{camelcase(name)}State')


@do(NvimIO[Any])
def component_state(component_name: str) -> Do:
    name = yield plugin_name()
    yield nvim_call_json(f'{camelcase(name)}ComponentState', component_name)


__all__ = ('program_log', 'seen_program', 'plugin_state', 'component_state',)



import inspect
from typing import Callable, Any, Tuple, get_type_hints

from amino import Maybe, _, Just, Boolean, Lists, Nothing, Either, L, List, Nil, Map, Left
from amino.dat import Dat
from amino.state import StateT
from amino.util.tpe import first_type_arg, type_arg, is_subclass

from ribosome.rpc.data.nargs import Nargs


def analyse_state_type(tpe: type) -> Tuple[Either[str, type], Either[str, type]]:
    return (
        (first_type_arg(tpe), type_arg(tpe, 1))
        if tpe is not None and is_subclass(tpe, StateT)
        else (Left('not a StateT'), Left('not a StateT'))
    )


def analyse_return_type(fun: Callable[..., Any], hints: Map[str, type]
                        ) -> Tuple[type, Either[str, type], Either[str, type]]:
    main_rettype = getattr(fun, 'tpe', hints.lift('return') | None)
    state_type, return_type = analyse_state_type(main_rettype)
    return main_rettype, state_type, return_type


def cons_params_spec(fun: Callable[..., Any]) -> None:
    argspec = inspect.getfullargspec(fun)
    hints = Map(get_type_hints(fun))
    params = Lists.wrap(argspec.args)
    defaults = Lists.wrap(argspec.defaults or ())
    method = Boolean(params.head.contains('self'))
    param_count = params.length - method.to_int
    min = param_count - defaults.length
    max = (~Boolean(argspec.varargs or argspec.varkw)).m(param_count)
    nargs = Nargs.cons(min, max)
    types = params.traverse(hints.lift, Maybe) | Nil
    main_rettype, state_type, return_type = analyse_return_type(fun, hints)
    return ParamsSpec(nargs, min, max, method, types, main_rettype, state_type, return_type | (lambda: main_rettype))


class ParamsSpec(Dat['ParamsSpec']):

    @staticmethod
    def from_function(fun: Callable[..., Any]) -> 'ParamsSpec':
        f = getattr(fun, '__wrapped__', fun)
        return cons_params_spec(f)

    @staticmethod
    def from_type(tpe: type) -> 'ParamsSpec':
        return cons_params_spec(tpe.__init__)

    def __init__(
            self,
            nargs: Nargs,
            min: int,
            max: Maybe[int],
            method: Boolean,
            types: List[type],
            rettype: type,
            state_type: Maybe[type],
            return_type: type,
    ) -> None:
        self.nargs = nargs
        self.min = min
        self.max = max
        self.method = method
        self.types = types
        self.rettype = rettype
        self.state_type = state_type
        self.return_type = return_type

    @property
    def exact_count(self) -> Maybe[int]:
        return Just(self.min) if self.max.contains(self.min) else Nothing


class ArgValidator(Dat['ArgValidator']):

    def __init__(self, spec: ParamsSpec) -> None:
        self.spec = spec

    @property
    def min(self) -> int:
        return self.spec.min

    @property
    def max(self) -> Maybe[int]:
        return self.spec.max

    def validate(self, count: int) -> Boolean:
        return Boolean(self.min <= count and not self.max.exists(_ < count))

    def error(self, args: tuple, desc: str, name: str) -> str:
        return f'argument count for {desc} `{name}` is {len(args)}, must be {self.count_spec} ({args})'

    @property
    def count_spec(self) -> str:
        return (
            self.spec.exact_count /
            (lambda a: f'exactly {a}' if a > 0 else 'none') | (
                self.max /
                (lambda mx: f'between {self.min} and {mx}') |
                f'at least {self.min}'
            )
        )

    def either(self, args: tuple, desc: str, name: str) -> Either[str, None]:
        return self.validate(len(args)).e(L(self.error)(args, desc, name), None)


__all__ = ('ArgValidator', 'ParamsSpec')



from typing import TypeVar

from amino import do, Do, _, __, List, Either, Maybe
from amino.lenses.lens import lens
from amino.state import EitherState
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PluginState
from ribosome.config.resolve import ComponentResolver
from ribosome.config.component import Components
from ribosome.nvim.api.command import nvim_command
from ribosome.nvim.io.compute import NvimIO
from ribosome.config import settings
from ribosome.rpc.define import define_rpc, ActiveRpcTrigger, undef_command
from ribosome.rpc.api import RpcProgram

log = module_log()
CC = TypeVar('CC')
D = TypeVar('D')
P = TypeVar('P')


def programs(state: PluginState[D, CC]) -> List[RpcProgram]:
    cfg_handlers = state.rpc
    compo_handlers = state.components.all // _.rpc
    return compo_handlers + cfg_handlers


@do(EitherState[str, PluginState[D, CC], None])
def update_components(requested: Either[str, List[str]]) -> Do:
    name = yield EitherState.inspect(_.basic.name)
    components_map = yield EitherState.inspect(_.comp.available)
    core_components = yield EitherState.inspect(_.basic.core_components)
    default_components = yield EitherState.inspect(_.basic.default_components)
    resolver = ComponentResolver(name, components_map, core_components, default_components, requested)
    components = yield EitherState.lift(resolver.run())
    yield EitherState.modify(__.copy(components=Components.cons(components)))
    progs = yield EitherState.inspect(programs)
    yield EitherState.modify(lens.programs.set(progs))


@do(NvimIO[None])
def undef_trigger(trigger: ActiveRpcTrigger) -> Do:
    yield nvim_command(undef_command(trigger.method), trigger.prog.rpc_name)


@do(NS[PluginState[D, CC], None])
def undef_triggers() -> Do:
    handlers = yield NS.inspect(_.rpc_triggers)
    yield NS.lift(handlers.traverse(undef_trigger, NvimIO))


@do(NS[PluginState[D, CC], None])
def def_triggers() -> Do:
    name = yield NS.inspect(_.basic.name)
    prefix = yield NS.inspect(_.basic.prefix)
    programs = yield NS.inspect(_.programs)
    handlers = yield NS.lift(define_rpc(programs, name, prefix))
    yield NS.modify(lens.rpc_triggers.set(handlers))


@do(NS[PluginState[D, CC], None])
def init_rpc(requested_components: Maybe[List[str]]) -> Do:
    yield update_components(requested_components).nvim
    yield def_triggers()


@do(NS[PluginState[D, CC], None])
def init_rpc_plugin() -> Do:
    requested_components = yield NS.lift(settings.components.value)
    yield init_rpc(requested_components)


__all__ = ('init_rpc', 'undef_triggers', 'def_triggers', 'programs', 'init_rpc_plugin',)



from amino import do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.command import nvim_command_output


@do(NvimIO[str])
def show_mappings() -> Do:
    lines = yield nvim_command_output('map')
    return lines.join_lines


__all__ = ('show_mappings',)



from typing import TypeVar, Callable

from lenses import UnboundLens

from amino import IO, Path, Left, do, Do, Either, Right
from amino.json import decode_json, dump_json
from amino.logging import module_log
from amino.util.fs import mkdir

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.state import NS
from ribosome.nvim.io.api import N
from ribosome.config.settings import project_state_dir

log = module_log()
A = TypeVar('A')
D = TypeVar('D')


@do(NvimIO[Path])
def check_state_dir(dir: Path, name: str) -> Do:
    yield N.from_io(mkdir(dir))
    yield N.pure(dir / f'{name}.json')


@do(NvimIO[Either[str, Path]])
def state_file(name: str) -> Do:
    dir = yield project_state_dir.value_or_default_e()
    yield dir.cata(lambda a: N.pure(Left(a)), lambda a: check_state_dir(a, name).map(Right))


@do(NvimIO[Either[str, A]])
def load_json_data_from(name: str, file: Path) -> Do:
    exists = yield N.from_io(IO.delay(file.exists))
    if exists:
        json = yield N.from_io(IO.delay(file.read_text))
        yield N.pure(decode_json(json))
    else:
        yield N.pure(Left(f'state file {file} does not exist'))


@do(NvimIO[Either[str, A]])
def load_json_data(name: str) -> Do:
    file = yield state_file(name)
    yield file.cata(lambda e: N.pure(Left(e)), lambda a: load_json_data_from(name, a))


@do(NS[D, None])
def load_json_state(name: str, store: UnboundLens) -> Do:
    state = yield NS.lift(load_json_data(name))
    yield state.cata(lambda a: NS.pure(None), lambda d: NS.modify(store.set(d)))


@do(NvimIO[None])
def store_json_to_file(name: str, data: A, file: Path) -> Do:
    json = yield N.from_either(dump_json(data))
    yield N.from_io(IO.delay(file.write_text, json))


@do(NvimIO[Either[str, None]])
def store_json_data(name: str, data: A) -> Do:
    file = yield state_file(name)
    yield file.cata(lambda e: N.pure(Left(e)), lambda a: store_json_to_file(name, data, a).map(Right))


@do(NS[D, Either[str, None]])
def store_json_state(name: str, fetch: Callable[[D], A]) -> Do:
    payload = yield NS.inspect(fetch)
    yield NS.lift(store_json_data(name, payload))


__all__ = ('load_json_state', 'store_json_data', 'store_json_state')



from amino.case import Case
from amino import Map
from amino.logging import module_log

from ribosome.nvim.syntax.expr import (SyntaxItem, SyntaxMatch, Highlight, HiLink, SyntaxKeyword, SyntaxRegion,
                                       SyntaxLiteral)

log = module_log()


def join_equals(data: Map[str, str]) -> str:
    return data.map2(lambda a, b: f'{a}={b}').join_tokens


class syntax_item_cmd(Case[SyntaxItem, str], alg=SyntaxItem):

    def keyword(self, data: SyntaxKeyword) -> str:
        options = data.options.join_tokens
        keywords = data.keywords.join_tokens
        params = join_equals(data.params)
        return f'syntax keyword {data.group} {data.keyword} {keywords} {options} {params}'

    def syntax_match(self, data: SyntaxMatch) -> str:
        options = data.options.join_tokens
        params = join_equals(data.params)
        return f'syntax match {data.group} /{data.pattern}/ {options} {params}'

    def region(self, data: SyntaxRegion) -> str:
        skip = data.skip.map(lambda a: f' skip=/{a}/').get_or_strict('')
        options = data.options.join_tokens
        params = join_equals(data.params)
        return f'syntax region {data.group} start=/{data.start}/{skip} end=/{data.end}/ {options} {params}'

    def literal(self, data: SyntaxLiteral) -> str:
        return data.cmd


def highlight_cmd(data: Highlight) -> str:
    values = join_equals(data.values)
    return f'highlight {data.group} {values}'


def hi_link_cmd(data: HiLink) -> str:
    return f'highlight link {data.group} {data.target}'


__all__ = ('syntax_item_cmd', 'highlight_cmd', 'hi_link_cmd',)



from concurrent.futures import Future
from typing import Callable, Tuple, Generic, TypeVar, Any

import msgpack

from amino import do, Do, IO, Right, Left, List, Lists
from amino.case import Case
from amino.logging import module_log

from ribosome.rpc.receive import (cons_receive, ReceiveResponse, ReceiveError, Receive, ReceiveRequest,
                                  ReceiveNotification, ReceiveExit, ReceiveUnknown)
from ribosome.rpc.comm import Comm
from ribosome.rpc.concurrency import Requests, RpcConcurrency
from ribosome.rpc.data.rpc import Rpc
from ribosome.rpc.data.rpc_type import BlockingRpc

log = module_log()
A = TypeVar('A')


def resolve_rpc(requests: Requests, id: int) -> IO[Tuple[Future, Rpc]]:
    return IO.delay(requests.to_vim.pop, id)


@do(IO[None])
def publish_response_from_vim(requests: Requests, data: ReceiveResponse) -> Do:
    fut = yield resolve_rpc(requests, data.id)
    yield IO.delay(fut.set_result, Right(data.data))


@do(IO[None])
def publish_error_from_vim(requests: Requests, err: ReceiveError) -> Do:
    fut = yield resolve_rpc(requests, err.id)
    yield IO.delay(fut.set_result, Left(err.error))


class handle_receive(Generic[A], Case[Receive, IO[None]], alg=Receive):

    def __init__(self, comm: Comm, execute_plugin_rpc: Callable[[Comm, Rpc], IO[None]]) -> None:
        self.comm = comm
        self.execute_plugin_rpc = execute_plugin_rpc

    @property
    def concurrency(self) -> RpcConcurrency:
        return self.comm.concurrency

    @property
    def requests(self) -> Requests:
        return self.concurrency.requests

    def response(self, data: ReceiveResponse) -> IO[None]:
        return self.concurrency.exclusive(publish_response_from_vim, self.requests, data)

    def error(self, err: ReceiveError) -> IO[None]:
        return self.concurrency.exclusive(publish_error_from_vim, self.requests, err)

    def request(self, receive: ReceiveRequest) -> IO[None]:
        return self.execute_plugin_rpc(self.comm, Rpc(receive.method, receive.args, BlockingRpc(receive.id)))

    def notification(self, receive: ReceiveNotification) -> IO[None]:
        return self.execute_plugin_rpc(self.comm, Rpc.nonblocking(receive.method, receive.args))

    def exit(self, receive: ReceiveExit) -> IO[None]:
        log.debug('exiting rpc session')
        return IO.pure(None)

    def unknown(self, receive: ReceiveUnknown) -> IO[None]:
        return IO.delay(log.error, f'received unknown rpc: {receive}')


@do(IO[List[Any]])
def unpack(data: bytes) -> Do:
    unpacker = msgpack.Unpacker()
    yield IO.delay(unpacker.feed, data)
    raw = yield IO.delay(list, unpacker)
    return Lists.wrap(raw)


def rpc_receive(comm: Comm, execute_plugin_rpc: Callable[[Comm, Rpc], None]) -> Callable[[bytes], IO[None]]:
    def handle(data: Any) -> IO[None]:
        receive = cons_receive(data)
        return handle_receive(comm, execute_plugin_rpc)(receive)
    @do(IO[None])
    def on_read(blob: bytes) -> Do:
        data = yield IO.delay(unpack, blob).recover_with(lambda e: IO.failed(f'failed to unpack: {e}'))
        yield data.traverse(handle, IO)
        yield IO.pure(None)
    return on_read


__all__ = ('rpc_receive',)



import sys
from typing import TypeVar, Any, Iterable

from lenses import UnboundLens

from amino.state import EitherState
from amino import do, Do, __, _, Map, Maybe, Lists, Just, Regex, L, IO, List
from amino.json import dump_json
from amino.dat import Dat
from amino.lenses.lens import lens
from amino.regex import Match
from amino.logging import module_log

from ribosome.compute.api import prog
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.io.state import NS
from ribosome.config.component import ComponentData
from ribosome.compute.prog import Prog
from ribosome.config.basic_config import NoData
from ribosome.components.internal.update import undef_triggers, def_triggers
from ribosome.compute.program import Program, bind_nullary_program
from ribosome.config.settings import run_internal_init
from ribosome.compute.ribosome_api import Ribo
from ribosome.rpc.data.rpc_method import RpcMethod
from ribosome.rpc.define import ActiveRpcTrigger
from ribosome.compute.output import Echo
from ribosome.nvim.api.rpc import plugin_name

log = module_log()
D = TypeVar('D')
CC = TypeVar('CC')


@prog
@do(NS[ComponentData[PluginState[D, CC], NoData], str])
def program_log() -> Do:
    yield NS.inspect_either(lambda s: dump_json(s.main.program_log))


@prog.unit
def set_log_level(level: str) -> NS[PluginState[D, CC], None]:
    handler = yield NS.inspect_f(_.file_log_handler)
    handler.setLevel(level)


@prog
@do(NS[ComponentData[PluginState[D, CC], NoData], str])
def state_data() -> Do:
    yield NS.inspect_either(lambda s: dump_json(s.main.data))


@prog
@do(NS[PluginState[D, CC], str])
def component_state_data(name: str) -> Do:
    data = yield NS.inspect_either(lambda s: s.data_by_name(name))
    yield NS.e(dump_json(data))


class RpcTrigger(Dat['RpcTrigger']):

    @staticmethod
    def from_active(trigger: ActiveRpcTrigger) -> 'RpcTrigger':
        return RpcTrigger(trigger.name, trigger.method)

    def __init__(self, name: str, method: RpcMethod) -> None:
        self.name = name
        self.method = method


@prog
def rpc_triggers() -> NS[PluginState[D, CC], List[RpcTrigger]]:
    return NS.inspect_either(lambda s: dump_json(s.rpc_triggers.map(RpcTrigger.from_active)))


class PatchQuery(Dat['PatchQuery']):

    def __init__(self, query: str, data: Map[str, Any]) -> None:
        self.query = query
        self.data = data


class UpdateQuery(Dat['UpdateQuery']):

    def __init__(self, patch: Maybe[PatchQuery]) -> None:
        self.patch = patch


find_rex = Regex(r'(?P<id>\w+)\((?P<key>\w+)=(?P<value>\w+)\)')


@do(Maybe[UnboundLens])
def find_lens(z: UnboundLens, m: Match) -> Do:
    id, key, value = yield m.all_groups('id', 'key', 'value').to_maybe
    yield Just(z.getattr_(id).each_().filter_(lambda a: Maybe.getattr(a, key).contains(value)))


# TODO `Regex.match_as(tpe)` that constructs a `Dat` from the `Match`
def lens_step(z: UnboundLens, s: str) -> Maybe[UnboundLens]:
    return find_rex.match(s) / L(find_lens)(z, _) | (lambda: Maybe.getattr(z, s))


def mk_lens(query: str) -> UnboundLens:
    return Lists.split(query, '.').fold_m(Just(lens))(lens_step)


@do(NS[PluginState[D, CC], None])
def patch_update(query: PatchQuery) -> Do:
    lns = yield NS.m(mk_lens(query.query), f'invalid state update query: {query.query}')
    lns1 = lens.data & lns
    yield NS.modify(lns1.modify(__.typed_copy(**query.data)))


@prog.unit
@do(NS[PluginState[D, CC], None])
def update_state(query: UpdateQuery) -> Do:
    yield query.patch / patch_update | NS.pure(None)


@do(NS[PluginState[D, CC], None])
def patch_update_component(comp: str, query: PatchQuery) -> Do:
    lns = yield NS.m(mk_lens(query.query), f'invalid component state update query for {comp}: {query.query}')
    yield NS.modify(__.modify_component_data(comp, lns.modify(__.typed_copy(**query.data))))


@prog.unit
@do(NS[PluginState[D, CC], None])
def update_component_state(comp: str, query: UpdateQuery) -> Do:
    yield query.patch / L(patch_update_component)(comp, _) | NS.pure(None)


@prog
def poll() -> NS[D, bool]:
    return NS.pure(True)


@prog.unit
def append_python_path(path: str) -> NS[D, None]:
    return NS.from_io(IO.delay(sys.path.append, path))


@prog
def show_python_path() -> NS[D, Iterable[str]]:
    return NS.pure(sys.path)


# FIXME need to update component type map
@prog.unit
@do(NS[PluginState[D, CC], None])
def enable_components(*names: str) -> Do:
    comps = (
        yield NS.inspect_either(
            __.comp.available.lift_all(*names).to_either(f'couldn\'t find some components: {names}'))
    )
    yield undef_triggers()
    yield NS.modify(lens.components.all.modify(__.add(comps)))
    yield def_triggers()
    print


class MapOptions(Dat['MapOptions']):

    @staticmethod
    def cons(
            buffer: int=None,
    ) -> 'MapOptions':
        return MapOptions(
            Maybe.optional(buffer),
        )

    def __init__(self, buffer: Maybe[int]) -> None:
        self.buffer = buffer


def mapping_data(ident: str, keys: str) -> NS[PluginState[D, CC], Program]:
    return NS.inspect_either(lambda a: a.active_mappings.lift(ident).to_either(f'no program for mapping `{keys}`'))


@prog
@do(NS[PluginState[D, CC], Program])
def mapping_program(ident: str, keys: str) -> Do:
    mapping, program = yield mapping_data(ident, keys)
    return program


@prog.do(None)
def mapping(ident: str, keys: str) -> Do:
    program = yield mapping_program(ident, keys)
    yield program()
    yield Prog.unit


@prog
@do(NS[PluginState[D, CC], Maybe[Program]])
def internal_init_program() -> Do:
    yield NS.inspect(_.init)


@prog.do(None)
def internal_init() -> Do:
    enabled = yield Ribo.setting_prog(run_internal_init)
    if enabled:
        program = yield internal_init_program()
        yield program / bind_nullary_program | Prog.unit


@prog.echo
@do(NS[D, Echo])
def rpc_job_stderr(id: int, data: list, event: str) -> Do:
    errors = Lists.wrap(data).filter(lambda a: len(a) > 0)
    name = yield NS.lift(plugin_name())
    log.debug(f'error in {name} rpc job on channel {id}:')
    errors.foreach(log.debug)
    msg = '' if len(errors) > 1 else errors.head.map(lambda a: f': {a}').get_or_strict('')
    yield NS.pure(Echo.error(f'fatal error in plugin {name}{msg}'))


__all__ = ('internal_init', 'mapping', 'MapOptions', 'enable_components', 'show_python_path', 'append_python_path',
           'poll', 'program_log', 'set_log_level', 'state_data', 'rpc_triggers', 'update_state',
           'update_component_state', 'component_state_data',)



from typing import TypeVar, Generic, Union, Callable

from kallikrein.matcher import BoundMatcher, StrictMatcher
from kallikrein.matchers.typed import ChainTyped, have_type
from kallikrein.matchers import contain
from kallikrein.matchers.contain import PredContain, NestContain
from kallikrein.match_result import MatchResult, ContainsMatchResult

from amino import Boolean
from amino.case import Case

from ribosome.nvim.io.data import NResult, NSuccess, NError, NFatal

A = TypeVar('A')
B = TypeVar('B')
Result = Union[A, str, Exception]


class apply_nresult(Generic[A, B], Case[NResult[A], B], alg=NResult):

    def __init__(self, target: Callable[[Result], B]) -> None:
        self.target = target

    def nsuccess(self, result: NSuccess[A]) -> B:
        return self.target(result.value)

    def nerror(self, result: NError[A]) -> B:
        return self.target(result.error)

    def nfatal(self, result: NFatal[A]) -> B:
        return self.target(result.exception)


class PredContainNResult(PredContain, tpe=NResult):

    def check(self, exp: NResult[A], target: Result) -> Boolean:
        def match(a: Result) -> Boolean:
            return Boolean(a == target)
        apply: apply_nresult[A, Boolean] = apply_nresult(match)
        return apply(exp)


class NestContainNResult(NestContain, tpe=NResult):

    def match(self, exp: NResult[A], target: BoundMatcher[Result]) -> MatchResult[Result]:
        apply: apply_nresult[A, MatchResult[Result]] = apply_nresult(target.evaluate)
        return apply(exp)

    def wrap(self, name: str, exp: NResult[A], nested: MatchResult[A]) -> MatchResult[A]:
        return ContainsMatchResult(name, exp, nested)


class ChainTypedNResult(ChainTyped, tpe=NResult):

    def chain(self, matcher: StrictMatcher, other: Union[A, BoundMatcher]) -> BoundMatcher:
        return matcher & contain(other)


nsuccess = have_type(NSuccess)
nerror = have_type(NError)
nfatal = have_type(NFatal)

__all__ = ('nsuccess', 'nerror', 'nfatal')



from numbers import Number
from typing import Any, Callable, TypeVar

from amino import Either, do, Do, Left, I
from amino.logging import module_log

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.rpc import plugin_name
from ribosome.nvim.api.util import cons_decode_str, cons_checked_e
from ribosome.nvim.api.data import Buffer
from ribosome.nvim.request import nvim_nonfatal_request, data_cons_request_nonfatal, nvim_request
from ribosome.nvim.io.api import N
from ribosome.nvim.api.exists import wait_until_valid

log = module_log()
A = TypeVar('A')


def variable_raw(name: str) -> NvimIO[Either[str, Any]]:
    return nvim_nonfatal_request('nvim_get_var', name)


@do(NvimIO[Either[str, Any]])
def variable_prefixed_raw(name: str) -> Do:
    plug = yield plugin_name()
    yield variable_raw(f'{plug}_{name}')


@do(NvimIO[Either[str, A]])
def variable(name: str, cons: Callable[[Any], Either[str, A]]) -> Do:
    log.debug(f'request variable `{name}`')
    def cons_e(result: Either[str, Any]) -> Either[str, A]:
        return result.cata(lambda err: Left(f'variable unset: {name}'), cons)
    value = yield data_cons_request_nonfatal('nvim_get_var', cons_e, name)
    log.debug(f'variable `{name}`: {value}')
    return value


def variable_str(name: str) -> NvimIO[Either[str, str]]:
    return variable(name, cons_decode_str)


def variable_num(name: str) -> NvimIO[Either[str, Number]]:
    return variable(name, cons_checked_e(Number, I))


@do(NvimIO[Either[str, A]])
def variable_prefixed(name: str, cons: Callable[[Any], Either[str, A]]) -> Do:
    plug = yield plugin_name()
    yield variable(f'{plug}_{name}', cons)


def variable_prefixed_str(name: str) -> NvimIO[Either[str, str]]:
    return variable_prefixed(name, cons_decode_str)


def variable_prefixed_num(name: str) -> NvimIO[Either[str, Number]]:
    return variable_prefixed(name, cons_checked_e(Number, I))


def variable_set(name: str, value: Any) -> NvimIO[None]:
    log.debug(f'setting variable `{name}` to `{value}`')
    return N.write('nvim_set_var', name, value)


@do(NvimIO[None])
def variable_set_prefixed(name: str, value: Any) -> Do:
    plug = yield plugin_name()
    ident = f'{plug}_{name}'
    log.debug(f'setting variable `{ident}` to `{value}`')
    yield N.write('nvim_set_var', ident, value)


def buffer_var_raw(buffer: Buffer, name: str) -> NvimIO[Any]:
    return nvim_request('nvim_buf_get_var', buffer.data, name)


def var_equals(target: Any) -> Callable[[str], NvimIO[bool]]:
    @do(NvimIO[bool])
    def var_equals(name: str) -> Do:
        actual = yield variable_raw(name)
        return actual.to_maybe.contains(target)
    return var_equals


def var_becomes(name: str, value: Any, timeout: float=3, interval: float=.01) -> NvimIO[None]:
    return wait_until_valid(name, var_equals(value), timeout=timeout, interval=interval, desc=f'become {value}')


@do(NvimIO[None])
def pvar_becomes(name: str, value: Any, timeout: float=3, interval: float=.01) -> Do:
    prefix = yield plugin_name()
    yield var_becomes(f'{prefix}_{name}', value, timeout=timeout, interval=interval)


__all__ = ('variable_raw', 'variable_prefixed_raw', 'variable', 'variable_prefixed', 'variable_prefixed_str',
           'variable_prefixed_num', 'variable_set', 'variable_set_prefixed', 'buffer_var_raw', 'variable_str',
           'variable_num')



import abc
from typing import TypeVar, Callable, Type

from amino.lenses.lens import lens
from amino import do, Do, _, Either

from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome import Ribosome
from ribosome.config.setting import Setting
from ribosome.nvim.io.compute import NvimIO, NRParams
from ribosome.compute.prog import Prog, ProgExec
from ribosome.compute.output import ProgOutputResult
from ribosome.compute.tpe_data import StateProg, trivial_state_prog, ribo_state_prog
from ribosome.compute.wrap import prog_wrappers
from ribosome.data.plugin_state import PS
from ribosome.nvim.api.command import doautocmd

A = TypeVar('A')
D = TypeVar('D')
M = TypeVar('M')
CC = TypeVar('CC')
C = TypeVar('C')
R = TypeVar('R')
main_lens = lens.state.data


class RMeta(abc.ABCMeta):
    pass


class Ribo(metaclass=RMeta):

    @classmethod
    def setting_e(self, setting: Setting[A]) -> NS[Ribosome[D, CC, C], Either[str, A]]:
        return NS.lift(setting.value_or_default_e())

    @classmethod
    def setting(self, setting: Setting[A]) -> NS[Ribosome[D, CC, C], A]:
        return NS.lift(setting.value_or_default())

    @classmethod
    def setting_raw(self, setting: Setting[A]) -> NS[Ribosome[D, CC, C], Either[str, A]]:
        return NS.lift(setting.value)

    @classmethod
    def setting_prog_e(self, setting: Setting[A]) -> Prog[Either[str, A]]:
        return Ribo.lift(Ribo.setting_e(setting), None)

    @classmethod
    def setting_prog(self, setting: Setting[A]) -> Prog[A]:
        return Ribo.lift(Ribo.setting(setting), None)

    @classmethod
    def comp(self) -> NS[Ribosome[D, CC, C], C]:
        return NS.inspect(lambda a: a.comp_lens.get()(a))

    @classmethod
    def inspect_comp(self, f: Callable[[C], A]) -> NS[Ribosome[D, CC, C], A]:
        return NS.inspect(lambda a: f(a.comp_lens.get()(a)))

    @classmethod
    def inspect_comp_e(self, f: Callable[[C], A]) -> NS[Ribosome[D, CC, C], A]:
        return NS.inspect_either(lambda a: f(a.comp_lens.get()(a)))

    @classmethod
    def modify_comp(self, f: Callable[[C], C]) -> NS[Ribosome[D, CC, C], None]:
        return NS.modify(lambda a: a.comp_lens.modify(f)(a))

    @classmethod
    def main(self) -> NS[Ribosome[D, CC, C], C]:
        return NS.inspect(lambda a: main_lens.get()(a))

    @classmethod
    def inspect_main(self, f: Callable[[D], A]) -> NS[Ribosome[D, CC, C], A]:
        return NS.inspect(lambda a: f(main_lens.get()(a)))

    @classmethod
    def modify_main(self, f: Callable[[D], D]) -> NS[Ribosome[D, CC, C], None]:
        return NS.modify(lambda a: main_lens.modify(f)(a))

    @classmethod
    @do(NS[Ribosome[D, CC, C], A])
    def zoom_main(self, fa: NS[D, A]) -> Do:
        yield fa.zoom(main_lens)

    @classmethod
    @do(NS[Ribosome[D, CC, C], A])
    def zoom_comp(self, fa: NS[C, A]) -> Do:
        lens = yield NS.inspect(_.comp_lens)
        yield fa.zoom(lens)

    @classmethod
    @do(Prog[A])
    def lift_state_prog(self, fa: NS[Ribosome[D, CC, C], A], state_type: StateProg[M, C, R, A]) -> Do:
        wrappers = yield Prog.from_either(prog_wrappers.match(state_type))
        yield ProgExec('lift', fa, wrappers, ProgOutputResult())

    @classmethod
    @do(Prog[A])
    def lift(self, fa: NS[Ribosome[D, CC, C], A], comp: Type[C]) -> Do:
        state_type: StateProg[PS[D, CC], C, Ribosome[D, CC, C]] = ribo_state_prog(comp)
        yield Ribo.lift_state_prog(fa, state_type)

    @classmethod
    def lift_comp(self, fa: NS[C, A], comp: Type[C]) -> Prog[A]:
        return Ribo.lift(Ribo.zoom_comp(fa), comp)

    @classmethod
    def trivial(self, fa: NS[D, A]) -> Prog[A]:
        return Ribo.lift_state_prog(fa, trivial_state_prog)

    @classmethod
    def lift_nvimio(self, fa: NvimIO[A]) -> Prog[A]:
        return Ribo.trivial(NS.lift(fa))

    @classmethod
    def autocmd(self, name: str) -> NS[Ribosome[D, CC, C], None]:
        return NS.lift(doautocmd('User', name, params=NRParams.cons(verbose=True, sync=False)))

    @classmethod
    def autocmd_prog(self, name: str) -> Prog[None]:
        return Ribo.lift(Ribo.autocmd(name), None)


__all__ = ('Ribosome', 'Ribo')



from typing import TypeVar, Callable, Generic, Tuple, cast

from amino.tc.base import F, ImplicitsMeta, Implicits
from amino import Either, List, options, Do, Boolean, Dat
from amino.state import State
from amino.do import do
from amino.dat import ADT, ADTMeta
from amino.case import Case, CaseRec, Term, RecStep
from amino.logging import module_log

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.trace import NvimIOException
from ribosome.nvim.io.data import NFatal, NResult, NSuccess, NError, Thunk, eval_thunk

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class NvimIOMeta(ImplicitsMeta, ADTMeta):
    pass


class NvimIO(
        Generic[A],
        Implicits,
        ADT['NvimIO[A]'],
        implicits=True,
        imp_mod='ribosome.nvim.io.tc',
        metaclass=NvimIOMeta,
):
    debug = options.io_debug.exists

    def eval(self) -> State[NvimApi, NResult[A]]:
        return eval_nvim_io(self)

    def run(self, vim: NvimApi) -> Tuple[NvimApi, NResult[A]]:
        try:
            return self.eval().run(vim).value
        except NvimIOException as e:
            return vim, NFatal(e)

    def run_s(self, vim: NvimApi) -> NvimApi:
        return self.run(vim)[0]

    def run_a(self, vim: NvimApi) -> NResult[A]:
        return self.run(vim)[1]

    result = run_a

    def drain(self, vim: NvimApi) -> None:
        self.run(vim)

    def either(self, vim: NvimApi) -> Either[NvimIOException, A]:
        return self.result(vim).to_either

    def unsafe_run(self, vim: NvimApi) -> Tuple[NvimApi, A]:
        result_vim, result = self.run(vim)
        return result_vim, result.to_either.get_or_raise()

    def unsafe(self, vim: NvimApi) -> A:
        return self.either(vim).get_or_raise()


class NvimIOPure(Generic[A], NvimIO[A]):

    def __init__(self, value: A) -> None:
        self.value = value


class NRParams(Dat['NRParams']):

    @staticmethod
    def cons(
            sync: bool=True,
            timeout: float=3.,
            decode: bool=True,
            verbose: bool=False,
    ) -> 'NRParams':
        return NRParams(
            sync,
            timeout,
            decode,
            verbose,
        )

    def __init__(self, sync: bool, timeout: float, decode: bool, verbose: bool) -> None:
        self.sync = sync
        self.timeout = timeout
        self.decode = decode
        self.verbose = verbose


class NvimIORequest(Generic[A], NvimIO[A]):

    def __init__(self, method: str, args: List[str], params: NRParams) -> None:
        self.method = method
        self.args = args
        self.params = params


class NvimIOSuspend(Generic[A], NvimIO[A]):

    @staticmethod
    def cons(thunk: State[NvimApi, NvimIO[A]]) -> 'NvimIOSuspend[A]':
        return NvimIOSuspend(Thunk.cons(thunk))

    def __init__(self, thunk: Thunk[NvimApi, NvimIO[A]]) -> None:
        self.thunk = thunk


class NvimIOBind(Generic[A, B], NvimIO[B]):

    def __init__(
            self,
            thunk: Thunk[NvimApi, NvimIO[A]],
            kleisli: Callable[[A], NvimIO[B]],
    ) -> None:
        self.thunk = thunk
        self.kleisli = kleisli


class NvimIOError(Generic[A], NvimIO[A]):

    def __init__(self, error: str) -> None:
        self.error = error


class NvimIOFatal(Generic[A], NvimIO[A]):

    def __init__(self, exception: Exception) -> None:
        self.exception = exception


class NvimIORecover(Generic[A], NvimIO[A]):

    def __init__(
            self,
            io: NvimIO[A],
            recover: Callable[[NResult[A]], NvimIO[A]],
            recoverable: Callable[[NResult[A]], Boolean],
    ) -> None:
        self.io = io
        self.recover = recover
        self.recoverable = recoverable


class lift_n_result(Case[NResult[A], NvimIO[A]], alg=NResult):

    def n_success(self, result: NSuccess[A]) -> NvimIO[A]:
        return NvimIOPure(result.value)

    def n_error(self, result: NError[A]) -> NvimIO[A]:
        return NvimIOError(result.error)

    def n_fatal(self, result: NFatal[A]) -> NvimIO[A]:
        return NvimIOFatal(result.exception)


@do(State[NvimApi, NvimIO[A]])
def execute_nvim_request(io: NvimIORequest[A]) -> Do:
    def make_request(vim: NvimApi) -> Tuple[NvimApi, NvimIO[A]]:
        response = vim.request(io.method, io.args, io.params.sync, io.params.timeout)
        updated_vim = response.map2(lambda v, a: v) | vim
        result = response.map2(lambda v, a: a)
        return updated_vim, NvimIOPure(result)
    updated_vim, result = yield State.inspect(make_request)
    yield State.reset(updated_vim, result)


@do(State[NvimApi, NvimIO[Either[List[str], A]]])
def execute_nvim_recover(io: NvimIORecover[A]) -> Do:
    def execute_wrapped(vim: NvimApi) -> Tuple[NvimApi, Either[List[str], A]]:
        updated_vim, result = io.io.run(vim)
        return updated_vim, (
            io.recover(result)
            if io.recoverable(result) else
            lift_n_result.match(result)
        )
    updated_vim, result = yield State.inspect(execute_wrapped)
    yield State.reset(updated_vim, result)


class flat_map_nvim_io(Case[Callable[[A], NvimIO[B]], NvimIO[B]], alg=NvimIO):

    def __init__(self, f: Callable[[A], NvimIO[B]]) -> None:
        self.f = f

    def nvim_io_pure(self, io: NvimIOPure[A]) -> NvimIO[B]:
        thunk = State.delay(self.f, io.value)
        return NvimIOSuspend.cons(thunk)

    def nvim_io_suspend(self, io: NvimIOSuspend[A]) -> NvimIO[B]:
        return NvimIOBind(io.thunk, self.f)

    def nvim_io_request(self, io: NvimIORequest[A]) -> NvimIO[B]:
        return NvimIOBind(Thunk.cons(execute_nvim_request(io)), self.f)

    def nvim_io_bind(self, io: NvimIOBind[C, A]) -> NvimIO[B]:
        return NvimIOSuspend.cons(State.pure(NvimIOBind(io.thunk, lambda a: io.kleisli(a).flat_map(self.f))))

    def nvim_io_error(self, io: NvimIOError[A]) -> NvimIO[B]:
        return cast(NvimIO[B], io)

    def nvim_io_fatal(self, io: NvimIOFatal[A]) -> NvimIO[B]:
        return cast(NvimIO[B], io)

    def nvim_io_recover(self, io: NvimIORecover[A]) -> NvimIO[B]:
        return NvimIOBind(Thunk.cons(execute_nvim_recover(io)), self.f)


EvalRes = RecStep[NvimIO[A], Tuple[NvimApi, NResult[A]]]


class eval_step(CaseRec[NvimIO[A], NResult[A]], alg=NvimIO):

    def __init__(self, vim: NvimApi) -> None:
        self.vim = vim

    def nvim_io_pure(self, io: NvimIOPure[A]) -> EvalRes:
        return Term((self.vim, NSuccess(io.value)))

    def nvim_io_request(self, io: NvimIORequest[A]) -> EvalRes:
        thunk = Thunk.cons(execute_nvim_request(io))
        return self(NvimIOSuspend(thunk))

    def nvim_io_suspend(self, io: NvimIOSuspend[A]) -> EvalRes:
        vim, next = eval_thunk(self.vim, io.thunk)
        return eval_step(vim)(next)

    def nvim_io_bind(self, io: NvimIOBind[B, A]) -> EvalRes:
        vim, next = eval_thunk(self.vim, io.thunk)
        return eval_step(vim)(next.flat_map(io.kleisli))

    def nvim_io_error(self, io: NvimIOError[A]) -> EvalRes:
        return Term((self.vim, NError(io.error)))

    def nvim_io_fatal(self, io: NvimIOFatal[A]) -> EvalRes:
        return Term((self.vim, NFatal(io.exception)))

    def nvim_io_recover(self, io: NvimIORecover[A]) -> NvimIO[B]:
        '''calls `map` to shift the recover execution to flat_map_nvim_io
        '''
        return eval_step(self.vim)(io.map(lambda a: a))


@do(State[NvimApi, A])
def eval_nvim_io(io: NvimIO[A]) -> Do:
    vim = yield State.get()
    updated_vim, result = eval_step(vim)(io).eval()
    yield State.set(updated_vim)
    return result


__all__ = ('NvimIO',)



import logging
from typing import TypeVar, Callable, Union

from toolz import merge

import amino
from amino.lazy import lazy
import amino.logging
from amino.logging import (amino_logger, init_loglevel, amino_root_file_logging, DDEBUG, print_log_info, VERBOSE,
                           LazyRecord, TEST, log_stamp, module_log)
from amino import Path, Logger, Nothing, Maybe, List, Lists

import ribosome  # noqa
from ribosome import options
from ribosome.nvim.api.ui import echo
from ribosome.util.string import escape_dquote

alog = module_log()
A = TypeVar('A')


def fmt_echo(text: Union[str, List[str]], cmd: str='echom', prefix: Maybe[str]=Nothing) -> List[str]:
    lines = text if isinstance(text, List) else Lists.lines(str(text))
    pre = prefix.map(_ + ': ') | ''
    return lines.map(lambda a: '{} "{}{}"'.format(cmd, pre, escape_dquote(a)))


def fmt_echohl(text: Union[str, List[str]], hl: str, prefix: Maybe[str]=Nothing) -> List[str]:
    return echo(text, prefix=prefix).cons(f'echohl {hl}').cat('echohl None')


class NvimHandler(logging.Handler):

    def __init__(self, vim: 'ribosome.NvimApi') -> None:
        self.vim = vim
        self.dispatchers = {
            logging.INFO: self.echo,
            logging.WARN: self.echowarn,
            logging.ERROR: self.echoerr,
            logging.CRITICAL: self.echoerr,
        }
        super().__init__()

    def emit(self, record: LazyRecord) -> None:
        dispatcher = self.dispatchers.get(record.levelno, self.echom)
        dispatcher(record.short())

    def echo(self, msg: str) -> None:
        echo(msg).unsafe(self.vim)

    def echowarn(self, msg: str) -> None:
        echo(msg).unsafe(self.vim)

    def echoerr(self, msg: str) -> None:
        echo(msg).unsafe(self.vim)

    def echom(self, msg: str) -> None:
        echo(msg).unsafe(self.vim)


class NvimFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> bool:
        return record.exc_info is None


nvim_filter = NvimFilter()
ribo_log = log = ribosome_root_logger = amino_logger('nvim')
_nvim_logging_initialized = False


def ribosome_logger(name: str) -> Logger:
    return ribosome_root_logger.getChild(name)


def ribosome_file_logging(name: str, file_kw: dict=dict()) -> logging.Handler:
    prefix_path = options.nvim_log_file.value | (lambda: amino.logging.log_dir() / 'nvim')
    level = (
        DDEBUG
        if options.development and options.spec else
        options.file_log_level.value | logging.DEBUG
    )
    logfile = Path(f'{prefix_path}_ribo_{name}_{log_stamp()}')
    kw = merge(
        file_kw,
        dict(level=level, logfile=logfile)
    )
    return amino_root_file_logging(**kw)


def ribosome_envvar_file_logging() -> None:
    fmt = options.file_log_fmt.value / (lambda fmt: dict(fmt=fmt)) | dict()
    options.ribo_log_file.value % (lambda f: amino_root_file_logging(logfile=Path(f), level=TEST, **fmt))


def ribosome_nvim_handler(vim: 'ribosome.NvimApi') -> None:
    handler = NvimHandler(vim)
    handler.addFilter(nvim_filter)
    ribosome_root_logger.addHandler(handler)
    init_loglevel(handler, VERBOSE)


def nvim_logging(vim: 'ribosome.NvimApi', file_kw: dict=dict()) -> logging.Handler:
    global _nvim_logging_initialized
    if not _nvim_logging_initialized:
        ribosome_nvim_handler(vim)
        _nvim_logging_initialized = True
        ribosome_envvar_file_logging()
        return ribosome_file_logging(vim.name, file_kw)


class Logging(amino.logging.Logging):

    @lazy
    def _log(self) -> Logger:
        return ribosome_logger(self.__class__.__name__)


def pr(a: A) -> A:
    v = log.verbose
    v(a)
    return a


def pv(a: A) -> A:
    v = log.verbose
    v(str(a))
    return a


def print_ribo_log_info(out: Callable[[str], None]) -> None:
    print_log_info(out)
    out(str(options.development))
    out(str(options.spec))
    out(str(options.file_log_level))
    out(str(options.file_log_fmt))
    out(str(options.nvim_log_file))

__all__ = ('ribosome_logger', 'nvim_logging', 'Logging', 'pr', 'ribo_log')



from typing import TypeVar, Tuple

from amino import List, IO, Dat, do, Do, Lists, Nil
from amino.case import Case
from amino.state import State
from amino.lenses.lens import lens

from ribosome.util.doc.data import (DocLine, DocCompiler, DocCat, DocFragment, DocString, DocBlock, DocMeta, NoMeta,
                                    Headline, Code, CustomDocMeta, Link)
from ribosome.util.doc.format import CompilerConfig, compile_anchor

A = TypeVar('A')


class VimCompilerConfig(Dat['VimCompilerConfig']):

    @staticmethod
    def cons(
            basic: CompilerConfig,
            width: int=78,
    ) -> 'VimCompilerConfig':
        return VimCompilerConfig(
            basic,
            width,
        )

    def __init__(self, basic: CompilerConfig, width: int) -> None:
        self.basic = basic
        self.width = width


def break_line(line: str, width: int) -> List[str]:
    tokens = Lists.tokens(line)
    def step(z: Tuple[List[str], str], a: str) -> Tuple[List[str], str]:
        lines, current = z
        return (
            (lines.cat(current), a)
            if len(current) + len(a) >= width else (
                (lines, f'{current} {a}')
                if len(current) > 0 else
                (lines, a)
            )
        )
    lines, rest = tokens.fold_left((Nil, ''))(step)
    return lines.cat(rest) if len(rest) > 0 else lines


@do(State[VimCompilerConfig, str])
def hline(char: str) -> Do:
    width = yield State.inspect(lambda a: a.width)
    return char * width


class compile_string(Case[DocMeta[A], State[VimCompilerConfig, List[str]]], alg=DocMeta):

    def __init__(self, string: DocString[A]) -> None:
        self.string = string

    @do(State[VimCompilerConfig, List[str]])
    def no(self, a: NoMeta[A]) -> Do:
        width = yield State.inspect(lambda a: a.width)
        return (
            List('')
            if len(self.string.text) == 0 else
            Lists.lines(self.string.text).flat_map(lambda a: break_line(a, width))
        )

    @do(State[VimCompilerConfig, List[str]])
    def headline(self, a: Headline[A]) -> Do:
        text = self.string.text
        width = yield State.inspect(lambda a: a.width)
        char = '=' if a.level == 1 else '-'
        hl = yield hline(char)
        name = yield State.inspect(lambda a: a.basic.name)
        anchor = yield (
            a.anchor
            .map(compile_anchor)
            .get_or(lambda: State.pure(f'{name}-{self.string.text.lower()}'))
            .zoom(lens.basic)
        )
        ws_len = width - (len(self.string.text) + len(anchor))
        ws = ' ' * ws_len
        main = text.upper() if a.level == 1 else text
        return List(hl, f'{main}{ws}*{anchor}*', '')

    def code(self, a: Code[A]) -> State[VimCompilerConfig, List[str]]:
        return State.pure(List(f'`{self.string.text}`'))

    def link(self, a: Link[A]) -> State[VimCompilerConfig, List[str]]:
        return State.pure(List(f'`[{a.desc}]({self.string.text})`'))

    def custom(self, a: CustomDocMeta[A]) -> State[VimCompilerConfig, List[str]]:
        return State.pure(List(self.string.text))


class compile_fragment(Case[DocFragment[A], List[str]], alg=DocFragment):

    @do(State[VimCompilerConfig, List[str]])
    def cat(self, frag: DocCat[A]) -> Do:
        parts = yield frag.fragments.flat_traverse(self, State)
        return List(parts.join_tokens)

    def string(self, frag: DocString[A]) -> State[VimCompilerConfig, List[str]]:
        return compile_string(frag)(frag.meta)


def compile_line(line: DocLine[A]) -> State[VimCompilerConfig, List[str]]:
    return compile_fragment.match(line.data)


def compile_block(block: DocBlock[A]) -> State[VimCompilerConfig, List[str]]:
    return block.lines.flat_traverse(compile_line, State)


@do(State[VimCompilerConfig, DocBlock[A]])
def footer() -> Do:
    hl = yield hline('=')
    width = yield State.inspect(lambda a: a.width)
    modeline = f'vim:tw={width}:ft=help:'
    return DocBlock.none(List(DocLine.string(hl), DocLine.string(modeline)))


def compile_vim(blocks: List[DocBlock[A]], conf: VimCompilerConfig) -> IO[List[str]]:
    foot = footer().run_a(conf).value
    return IO.pure(blocks.cat(foot).flat_traverse(compile_block, State).run_a(conf).value)


def vim_compiler(conf: VimCompilerConfig) -> DocCompiler[A, VimCompilerConfig]:
    return DocCompiler(compile_vim, conf)


__all__ = ('compile_vim', 'vim_compiler',)



import time
from typing import Callable, TypeVar, Type, Any, List as TList, Tuple

from msgpack import ExtType

from amino import Either, Left, Right, List, Lists, Try, do, Maybe, Do, Just, Nothing
from amino.json import decode_json
from amino.json.decoder import decode_json_type

from ribosome.nvim.io.compute import NvimIO, lift_n_result
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NResult, NSuccess

A = TypeVar('A')
B = TypeVar('B')


def cons_checked_e(tpe: Type[A], cons: Callable[[A], Either[str, B]]) -> Callable[[Any], Either[str, B]]:
    def cons_data(data: Any) -> Either[str, A]:
        return cons(data) if isinstance(data, tpe) else Left(f'invalid nvim data for `{tpe}`: {data}')
    return cons_data


def cons_checked(tpe: Type[A], cons: Callable[[A], B]) -> Callable[[Any], Either[str, B]]:
    return cons_checked_e(tpe, lambda a: Right(cons(a)))


def cons_ext(cons: Callable[[ExtType], A]) -> Callable[[Any], Either[str, A]]:
    return cons_checked(ExtType, cons)


def cons_checked_list_e(tpe: Type[A], cons: Callable[[A], Either[str, B]]) -> Callable[[Any], Either[str, List[B]]]:
    def cons_data(data: Any) -> Either[str, List[B]]:
        return (
            Lists.wrap(data).traverse(cons_checked_e(tpe, cons), Either)
            if isinstance(data, TList) else
            Left(f'invalid nvim data for `List[{tpe}]`: {data}')
        )
    return cons_data


def cons_checked_list(tpe: Type[A], cons: Callable[[A], B]) -> Callable[[Any], Either[str, List[B]]]:
    return cons_checked_list_e(tpe, lambda a: Right(cons(a)))


def cons_ext_list(cons: Callable[[ExtType], A]) -> Callable[[Any], Either[str, List[A]]]:
    return cons_checked_list(ExtType, cons)


cons_decode_str = cons_checked_e((bytes, str), (lambda a: Try(a.decode) if isinstance(a, bytes) else Right(a)))


def check_str_list(a: list) -> Either[str, List[str]]:
    return Lists.wrap(a).traverse(cons_decode_str, Either)


cons_decode_str_list = cons_checked_e(list, check_str_list)


cons_split_lines = cons_checked(str, Lists.lines)


def check_bool(a: Any) -> Either[str, bool]:
    return Right(a == 1) if a == 1 or a == 0 else Left(f'nvim result is not a boolean: {a}')


cons_decode_bool = cons_checked_e(int, check_bool)


def extract_int_pair(data: list) -> Either[str, Tuple[int, int]]:
    data_l = Lists.wrap(data)
    @do(Maybe[Tuple[int, int]])
    def run() -> Do:
        first = yield data_l.lift(0)
        second = yield data_l.lift(1)
        yield Just((first, second)) if isinstance(first, int) and isinstance(second, int) else Nothing
    return run().to_either(lambda: f'invalid nvim data for int pair: {data}')


def split_option(value: str) -> List[str]:
    return Lists.split(value, ',')


cons_decode_str_list_option = cons_checked(str, split_option)

cons_json = cons_checked_e(str, decode_json)

cons_int = cons_checked_e(int, lambda a: Right(a))


def cons_json_tpe(tpe: Type[A]) -> NvimIO[A]:
    return cons_checked_e(str, lambda a: decode_json_type(a, tpe))


@do(NvimIO[A])
def nvimio_repeat_timeout(
        thunk: Callable[[], NvimIO[A]],
        check: Callable[[A], bool],
        error: str,
        timeout: float,
        interval: float=None,
) -> Do:
    effective_interval = .01 if interval is None else interval
    start = yield N.simple(time.time)
    @do(NvimIO[None])
    def wait_and_recurse() -> Do:
        yield N.sleep(effective_interval)
        yield recurse()
    @do(NvimIO[None])
    def recurse() -> Do:
        result = yield thunk()
        done = check(result)
        yield (
            N.pure(result)
            if done else
            N.error(error)
            if time.time() - start > timeout else
            wait_and_recurse()
        )
    yield recurse()


def nvimio_result(thunk: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> NvimIO[NResult[A]]:
    return N.intercept(thunk(*a, **kw), N.pure)


@do(NvimIO[A])
def nvimio_await_success(
        thunk: Callable[..., NvimIO[A]],
        timeout: float,
        *a: Any,
        interval: float=.01,
        **kw: Any,
) -> Do:
    start = yield N.simple(time.time)
    @do(NvimIO[None])
    def wait_and_recurse() -> Do:
        yield N.sleep(interval)
        yield recurse()
    @do(NvimIO[None])
    def recurse() -> Do:
        result = yield nvimio_result(thunk, *a, **kw)
        yield (
            lift_n_result.match(result)
            if isinstance(result, NSuccess) or time.time() - start > timeout else
            wait_and_recurse()
        )
    yield recurse()


def format_bufdo(cmd: str, start: int, end: Maybe[int]) -> str:
    range = end.map(lambda a: f'{start},{a}').get_or(str, start)
    return f'{range}bufdo {cmd}'


def format_windo(cmd: str, start: int, end: Maybe[int]) -> str:
    range = end.map(lambda a: f'{start},{a}').get_or(str, start)
    return f'{range}windo {cmd}'


__all__ = ('cons_checked_e', 'cons_checked', 'cons_ext', 'cons_checked_list', 'cons_ext_list', 'check_str_list',
           'cons_decode_str', 'cons_decode_str_list', 'extract_int_pair', 'split_option', 'cons_decode_str_list_option',
           'cons_split_lines', 'cons_decode_bool', 'nvimio_repeat_timeout', 'cons_json', 'cons_json_tpe',
           'nvimio_await_success', 'cons_checked_list_e', 'format_bufdo', 'format_windo', 'cons_int',)



from amino import do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.exists import command_exists, command_exists_not

from kallikrein import Expectation, k


@do(NvimIO[Expectation])
def command_must_exist(name: str) -> Do:
    exists = yield command_exists(name)
    return k(exists).true


@do(NvimIO[Expectation])
def command_must_not_exist(name: str) -> Do:
    exists_not = yield command_exists_not(name)
    return k(exists_not).true


__all__ = ('command_must_exist', 'command_must_not_exist',)



from amino import Dat, Map, ADT, List, Nil, Maybe, Lists


class SyntaxItem(ADT['SyntaxItem']):
    pass


class SyntaxKeyword(SyntaxItem):

    @staticmethod
    def cons(
            group: str,
            keyword: str,
            *options: str,
            keywords: List[str]=Nil,
            **params: str,
    ) -> 'SyntaxKeyword':
        return SyntaxKeyword(
            group,
            keyword,
            keywords,
            Lists.wrap(options),
            Map(params),
        )

    def __init__(
            self,
            group: str,
            keyword: str,
            keywords: List[str],
            options: List[str],
            params: Map[str, str],
    ) -> None:
        self.group = group
        self.keyword = keyword
        self.keywords = keywords
        self.params = params
        self.options = options
        self.params = params


class SyntaxMatch(SyntaxItem):

    @staticmethod
    def cons(
            group: str,
            pattern: str,
            *options: str,
            **params: str,
    ) -> 'SyntaxMatch':
        return SyntaxMatch(
            group,
            pattern,
            Lists.wrap(options),
            Map(params),
        )

    def __init__(self, group: str, pattern: str, options: List[str], params: Map[str, str]) -> None:
        self.group = group
        self.pattern = pattern
        self.options = options
        self.params = params


class SyntaxRegion(SyntaxItem):

    @staticmethod
    def cons(
            group: str,
            start: str,
            end: str,
            *options: str,
            skip: str=None,
            starts: List[str]=Nil,
            stops: List[str]=Nil,
            **params: str,
    ) -> 'SyntaxRegion':
        return SyntaxRegion(
            group,
            start,
            end,
            Maybe.optional(skip),
            Lists.wrap(options),
            Map(params),
        )

    def __init__(
            self,
            group: str,
            start: str,
            end: str,
            skip: Maybe[str],
            options: List[str],
            params: Map[str, str],
    ) -> None:
        self.group = group
        self.start = start
        self.end = end
        self.skip = skip
        self.options = options
        self.params = params


class SyntaxLiteral(SyntaxItem):

    def __init__(self, cmd: str) -> None:
        self.cmd = cmd


class Highlight(Dat['Highlight']):

    @staticmethod
    def cons(
            group: str,
            **values: int,
    ) -> 'Highlight':
        return Highlight(
            group,
            Map(values),
        )

    def __init__(self, group: str, values: Map[str, int]) -> None:
        self.group = group
        self.values = values


class HiLink(Dat['HiLink']):

    @staticmethod
    def cons(
            group: str,
            target: str,
    ) -> 'HiLink':
        return HiLink(
            group,
            target,
        )

    def __init__(self, group: str, target: str) -> None:
        self.group = group
        self.target = target


__all__ = ('SyntaxMatch',)



from amino import ADT


class RpcReadError(ADT['RpcReadError']):
    pass


class RpcProcessExit(RpcReadError):
    pass


class RpcReadErrorUnknown(RpcReadError):

    def __init__(self, message: str) -> None:
        self.message = message


__all__ = ('RpcReadError', 'RpcProcessExit', 'RpcReadErrorUnknown')



from typing import Any

from amino import do, Do, IO, List

from amino.logging import module_log

from ribosome.rpc.comm import Comm, Exec
from ribosome.rpc.to_vim import handle_response, RpcResponse
from ribosome.rpc.nvim_api import RiboNvimApi
from ribosome.rpc.data.rpc import Rpc
from ribosome.nvim.io.data import NFatal, NResult
from ribosome.rpc.response import validate_rpc_result, report_error

log = module_log()


# TODO fetch nvim api ctor from `Comm`
@do(IO[NResult[List[Any]]])
def execute_rpc(comm: Comm, rpc: Rpc, execute: Exec, plugin_name: str) -> Do:
    thunk = yield IO.delay(execute, rpc.method, rpc.args)
    yield IO.delay(thunk.run_a, RiboNvimApi(plugin_name, comm))


@do(IO[RpcResponse])
def execute_rpc_safe(comm: Comm, rpc: Rpc, execute: Exec, plugin_name: str) -> Do:
    result = yield execute_rpc(comm, rpc, execute, plugin_name).recover(NFatal)
    validated = validate_rpc_result(rpc)(result)
    yield report_error.match(validated)
    return validated


@do(IO[Any])
def execute_rpc_from_vim(rpc: Rpc, comm: Comm, execute: Exec, plugin_name: str) -> Do:
    result = yield execute_rpc_safe(comm, rpc, execute, plugin_name)
    yield handle_response.match(result).run(comm)


__all__ = ('execute_rpc_from_vim')



from typing import Callable, TypeVar
from threading import Event
from queue import Queue

from amino import do, Do, IO

from ribosome.nvim.io.compute import NvimIO, lift_n_result
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NResult, NError
from ribosome.util.menu.prompt.data import PromptInterrupt, InputResources
from ribosome.nvim.io.state import NS

A = TypeVar('A')
B = TypeVar('B')


@do(IO[None])
def stop_prompt(queue: Queue, stop: Event) -> Do:
    yield IO.delay(stop.set)
    yield IO.delay(queue.put, PromptInterrupt())


@do(NS[InputResources[A, B], None])
def stop_prompt_s() -> Do:
    res = yield NS.get()
    yield NS.lift(N.from_io(stop_prompt(res.inputs, res.stop)))


def intercept_interrupt_result(queue: Queue, stop: Event, default: A) -> Callable[[NResult[A]], NvimIO[A]]:
    @do(NvimIO[A])
    def intercept_interrupt_result(result: NResult[A]) -> Do:
        yield N.from_io(stop_prompt(queue, stop))
        yield (
            N.pure(default)
            if isinstance(result, NError) and 'Keyboard interrupt' in str(result.error) else
            lift_n_result.match(result)
        )
    return intercept_interrupt_result


def intercept_interrupt(queue: Queue, stop: Event, default: A, thunk: NvimIO[A]) -> NvimIO[A]:
    return N.recover_failure(thunk, intercept_interrupt_result(queue, stop, default))


__all__ = ('stop_prompt', 'intercept_interrupt', 'stop_prompt_s',)



from typing import Any, Callable, TypeVar, Generic, Optional
from threading import Lock

from amino import Dat, List, Either, IO, do, Do, Try
from amino.logging import module_log

from ribosome.nvim.io.api import N
from ribosome.nvim.io.state import NS
from ribosome.nvim.io.compute import NvimIO, lift_n_result
from ribosome.nvim.io.data import NResult
from ribosome.rpc.concurrency import RpcConcurrency, OnMessage, OnError
from ribosome.rpc.data.rpc import Rpc

A = TypeVar('A')
B = TypeVar('B')
log = module_log()
Exec = Callable[[str, List[Any]], NvimIO[List[Any]]]


# FIXME send/exit should be IO
class RpcComm(Dat['RpcComm']):

    def __init__(
            self,
            start_processing: Callable[[OnMessage, OnError], IO[None]],
            stop_processing: Callable[[], IO[None]],
            send: Callable[[bytes], None],
            join: Callable[[], IO[None]],
            exit: Callable[[], None],
    ) -> None:
        self.start_processing = start_processing
        self.stop_processing = stop_processing
        self.send = send
        self.join = join
        self.exit = exit


class Comm(Dat['Comm']):

    @staticmethod
    def cons(
            request_handler: Exec,
            rpc: RpcComm,
            concurrency: RpcConcurrency=None,
    ) -> 'Comm':
        return Comm(request_handler, rpc, concurrency or RpcConcurrency.cons())

    def __init__(
            self,
            request_handler: Callable[[str, List[Any], bool], Either[str, Any]],
            rpc: RpcComm,
            concurrency: RpcConcurrency,
    ) -> None:
        self.request_handler = request_handler
        self.rpc = rpc
        self.concurrency = concurrency

    @property
    def lock(self) -> Lock:
        return self.concurrency.lock


class StateGuard(Generic[A], Dat['StateGuard']):

    @staticmethod
    def cons(state: A, initialized: bool=False) -> 'StateGuard[A]':
        return StateGuard(state, initialized, Lock())

    def __init__(self, state: A, initialized: bool, lock: Lock) -> None:
        self.state = state
        self.initialized = initialized
        self.lock = lock

    def exclusive(self, f: Callable[..., A], *a: Any, **kw: Any) -> Any:
        with self.lock:
            return f(*a, **kw)

    def exclusive_state(self, f: Callable[..., A], *a: Any, **kw: Any) -> None:
        with self.lock:
            self.state = f(self.state, *a, **kw)

    def update(self, state: A) -> None:
        self.exclusive(setattr, self, 'state', state)

    def init(self, state: A) -> None:
        self.update(state)
        self.exclusive(setattr, self, 'initialized', True)

    @do(NvimIO[None])
    def acquire(self) -> Do:
        yield N.simple(self.lock.acquire)

    @do(NvimIO[None])
    def release(self, result: Optional[NResult[A]]=None) -> Do:
        yield N.simple(Try, self.lock.release)
        if result:
            log.debug(f'released lock due to error: {result}')
            yield lift_n_result.match(result)


def unsafe_update_state(guard: StateGuard[A], state: A) -> None:
    guard.state = state


@do(NvimIO[B])
def exclusive_ns(guard: StateGuard[A], desc: str, thunk: Callable[..., NS[A, B]], *a: Any) -> Do:
    '''this is the central unsafe function, using a lock and updating the state in `guard` in-place.
    '''
    yield guard.acquire()
    log.debug2(lambda: f'exclusive: {desc}')
    state, response = yield N.ensure_failure(thunk(*a).run(guard.state), guard.release)
    yield N.delay(lambda v: unsafe_update_state(guard, state))
    yield guard.release()
    log.debug2(lambda: f'release: {desc}')
    yield N.pure(response)


__all__ = ('Comm', 'OnError', 'RpcComm', 'StateGuard')



from amino import ADT


class RpcMethod(ADT['RpcMethod']):
    pass


class CommandMethod(RpcMethod):

    @staticmethod
    def cons(bang: bool=False) -> 'CommandMethod':
        return CommandMethod(bang)

    def __init__(self, bang: bool) -> None:
        self.bang = bang


class FunctionMethod(RpcMethod):

    @property
    def method(self) -> str:
        return 'function'


class AutocmdMethod(RpcMethod):

    @staticmethod
    def cons(pattern: str=None, sync: bool=False) -> 'AutocmdMethod':
        return AutocmdMethod(pattern or '*', sync)

    def __init__(self, pattern: str, sync: bool) -> None:
        self.pattern = pattern
        self.sync = sync


__all__ = ('RpcMethod', 'CommandMethod', 'FunctionMethod', 'AutocmdMethod',)



from typing import TypeVar, Callable, Any
from concurrent.futures import wait, ThreadPoolExecutor

from amino.case import Case
from amino import IO, List, Either, Lists, __, Maybe, Nothing, do, Do
from amino.io import IOException

from ribosome import ribo_log
from ribosome.compute.output import (ProgOutput, ProgIO, ProgOutputIO, ProgScalarIO, ProgGatherIOs,
                                     ProgScalarSubprocess, ProgGatherSubprocesses, GatherIOs, GatherSubprocesses,
                                     ProgIOCustom, ProgOutputUnit, ProgOutputResult, ProgIOEcho, Echo, GatherItem,
                                     GatherIO, GatherSubprocess, GatherResult, GatherIOResult, GatherSubprocessResult,
                                     Gather, ProgGather)
from ribosome.compute.prog import Prog
from ribosome.compute.program import Program
from ribosome.process import Subprocess
from ribosome.nvim.io.state import NS

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')


def gather_ios(gio: GatherIOs[A]) -> List[Either[IOException, A]]:
    with ThreadPoolExecutor(thread_name_prefix='ribosome_dio') as executor:
        ribo_log.debug(f'executing ios {gio.ios}')
        futures = gio.ios.map(lambda i: executor.submit(i.attempt_run))
        completed, timed_out = wait(futures, timeout=gio.timeout)
        ribo_log.debug(f'completed ios {completed}')
        if timed_out:
            ribo_log.debug(f'ios timed out: {timed_out}')
        return Lists.wrap(completed).map(__.result(timeout=gio.timeout))


def gather_subprocesses(gio: GatherSubprocesses[A]) -> List[Either[IOException, A]]:
    ribo_log.debug(f'gathering {gio}')
    popens = gio.subprocs.map(__.execute())
    return gather_ios(GatherIOs(popens, gio.timeout))


class gather_item(Case[GatherItem[A], IO[GatherResult[A]]], alg=GatherItem):

    def io(self, a: GatherIO[A]) -> IO[GatherResult[A]]:
        return a.io.map(GatherIOResult)

    def subproc(self, a: GatherSubprocess[A]) -> IO[GatherResult[A]]:
        return a.subprocess.execute().map(GatherSubprocessResult)


def gather_mixed(gio: Gather[A]) -> List[Either[IOException, GatherSubprocessResult[A]]]:
    ios = gio.items.map(gather_item.match)
    return gather_ios(GatherIOs(ios, gio.timeout))


ProgIOInterpreter = Callable[[ProgIO, Any], Prog[A]]


def default_logger(echo: Echo) -> NS[D, None]:
    io = echo.messages.traverse(lambda m: IO.delay(ribo_log.log, echo.level, m), IO)
    return NS.from_io(io).replace(None)


default_logger_program = Program.lift(default_logger)


class interpret_io(Case[ProgIO, Prog[A]], alg=ProgIO):

    def __init__(self, custom: Callable[[Any], Prog[A]], logger: Maybe[Program]=Nothing) -> None:
        self.custom = custom
        self.logger = logger

    def prog_scalar_io(self, po: ProgScalarIO, output: IO[A]) -> Prog[A]:
        return Prog.from_either(output.attempt)

    def prog_gather_ios(self, po: ProgGatherIOs, output: GatherIOs[A]) -> Prog[Prog[List[Either[IOException, A]]]]:
        return Prog.pure(gather_ios(output))

    def prog_scalar_subprocess(self, po: ProgScalarSubprocess, output: Subprocess[A]) -> Prog[A]:
        return Prog.from_either(output.execute().attempt)

    def prog_gather_subprocesses(self, po: ProgGatherSubprocesses, output: GatherSubprocesses[A]
                                 ) -> Prog[List[Either[IOException, A]]]:
        return Prog.pure(gather_subprocesses(output))

    @do(Prog[None])
    def prog_io_echo(self, po: ProgIOEcho, output: Echo) -> Do:
        logger = self.logger | (lambda: default_logger_program)
        yield logger(output).replace(None)

    def prog_io_custom(self, po: ProgIOCustom, output: Any) -> Prog[A]:
        return self.custom(output)

    def prog_gather(self, po: ProgGather, output: Gather[A]) -> Prog[List[Either[IOException, A]]]:
        return Prog.pure(gather_mixed(output))


class interpret(Case[ProgOutput, Prog[B]], alg=ProgOutput):

    def __init__(self, io: ProgIOInterpreter[B]) -> None:
        self.io = io

    def prog_output_io(self, po: ProgOutputIO, output: IO[B]) -> Prog[B]:
        return self.io(po.io, output)

    def prog_output_unit(self, po: ProgOutputUnit, output: Any) -> Prog[None]:
        return Prog.pure(None)

    def prog_output_result(self, po: ProgOutputResult, output: B) -> Prog[B]:
        return Prog.pure(output)


def no_interpreter(po: ProgOutputIO, a: Any) -> Prog[A]:
    return Prog.error(f'no custom interpreter ({a})')


__all__ = ('ProgIOInterpreter', 'interpret_io', 'interpret', 'no_interpreter')



from __future__ import annotations
import abc
from typing import Any, Callable, Tuple

from msgpack import ExtType

from amino import List, Either, Map, Left, Dat, Nil, do, Do
from amino.logging import module_log

log = module_log()


class NvimApi(Dat['NvimApi']):

    def __init__(self, name: str) -> None:
        self.name = name

    @abc.abstractmethod
    def request(self, method: str, args: List[Any], sync: bool, timeout: float) -> Either[str, Tuple['NvimApi', Any]]:
        ...


StrictNvimHandler = Callable[['StrictNvimApi', str, List[Any], bool], Either[List[str], Tuple[NvimApi, Any]]]


def no_request_handler(vim: 'StrictNvimApi', method: str, args: List[Any], sync: bool
                       ) -> Either[List[str], Tuple[NvimApi, Any]]:
    return Left(List('no request handler defined'))


class StrictNvimApi(NvimApi):

    @staticmethod
    def cons(
            name: str,
            vars: Map[str, Any]=Map(),
            request_handler: StrictNvimHandler=no_request_handler,
            request_log: List[Tuple[str, List[Any]]]=Nil,
    ) -> 'StrictNvimApi':
        return StrictNvimApi(name, vars, request_handler, request_log)

    def __init__(
            self,
            name: str,
            vars: Map[str, Any],
            request_handler: StrictNvimHandler,
            request_log: List[Tuple[str, List[Any]]],
    ) -> None:
        self.name = name
        self.vars = vars
        self.request_handler = request_handler
        self.request_log = request_log

    def request(self, method: str, args: List[Any], sync: bool, timeout: float) -> Either[str, Tuple[NvimApi, Any]]:
        vim = self.append1.request_log((method, args))
        return self.request_handler(vim, method, args, sync).accum_error_lift(variable_request, vim, method, args)

    def var(self, name: str) -> Either[str, Any]:
        return self.vars.lift(name).to_either(f'no variable `{name}` defined')


@do(Either[str, Tuple[NvimApi, Any]])
def get_var(vim: StrictNvimApi, name: str) -> Do:
    value = yield vim.var(name)
    return (vim, value)


@do(Either[str, Tuple[NvimApi, Any]])
def set_var(vim: StrictNvimApi, name: str, rest: List[Any]) -> Do:
    value, rest1 = yield rest.uncons.to_either(f'no value specified for updating var `{name}`')
    return vim.append.vars((name, value)), None


def manipulate_vars(vim: StrictNvimApi, method: str, name: str, rest: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return (
        get_var(vim, name)
        if method == 'nvim_get_var' else
        set_var(vim, name, rest)
        if method == 'nvim_set_var' else
        Left(f'not a variable request')
    )


@do(Either[str, Tuple[NvimApi, Any]])
def variable_request(vim: StrictNvimApi, method: str, args: List[Any]) -> Do:
    name, rest = yield args.uncons.to_either('no variable name given')
    yield manipulate_vars(vim, method, name, rest)


class Tabpage(Dat['Tabpage']):

    def __init__(self, data: ExtType) -> None:
        self.data = data


class Window(Dat['Window']):

    def __init__(self, data: ExtType) -> None:
        self.data = data


class Buffer(Dat['Buffer']):

    def __init__(self, data: ExtType) -> None:
        self.data = data


__all__ = ('NvimApi', 'StrictNvimApi', 'Tabpage', 'Window', 'StrictNvimHandler')



from typing import Any, Callable, TypeVar

from amino import List, do, Do, Nil
from amino.logging import module_log
from amino.util.string import decode

from ribosome.nvim.io.compute import NvimIO
from ribosome.rpc.comm import StateGuard, exclusive_ns
from ribosome.nvim.io.api import N
from ribosome.compute.program import Program
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PS
from ribosome.compute.run import run_prog
from ribosome.compute.api import parse_args
from ribosome.rpc.api import RpcProgram
from ribosome.rpc.data.rpc import RpcArgs
from ribosome.nvim.api.util import nvimio_repeat_timeout

log = module_log()
A = TypeVar('A')


@do(NS[PS, Any])
def run_program(rpc_program: RpcProgram, args: RpcArgs) -> Do:
    parsed_args = yield NS.from_either(parse_args(rpc_program, args.args))
    yield run_prog(rpc_program.program, parsed_args)


def run_programs(programs: List[Program], args: RpcArgs) -> NvimIO[List[Any]]:
    return programs.traverse(lambda a: run_program(a, args), NvimIO)


def run_program_exclusive(guard: StateGuard[A], program: RpcProgram, args: RpcArgs) -> NvimIO[Any]:
    return (
        exclusive_ns(guard, program.program.name, run_program, program, args)
        if program.options.write else
        run_program(program, args).run_a(guard.state)
    )


def run_programs_exclusive(guard: StateGuard[A], programs: List[Program], args: RpcArgs) -> NvimIO[List[Any]]:
    return programs.traverse(lambda a: run_program_exclusive(guard, a, args), NvimIO)


def no_programs_for_rpc(method: str, args: RpcArgs) -> NvimIO[A]:
    return N.error(f'no programs defined for request {method}({args.string})')


def decode_args(args: List[Any]) -> RpcArgs:
    decoded_args = decode(args)
    fun_args = decoded_args.head | Nil
    bang = decoded_args.lift(1).contains(1)
    return RpcArgs(fun_args, bang)


def rpc_handler(guard: StateGuard[A]) -> Callable[[str, List[Any]], NvimIO[List[Any]]]:
    @do(NvimIO[List[Any]])
    def handler(method: str, raw_args: List[Any]) -> Do:
        yield nvimio_repeat_timeout(lambda: N.pure(guard), lambda a: a.initialized, '''state wasn't initialized''', 20)
        args = decode_args(raw_args)
        log.debug(f'handling request: {method}({args.args.join_comma})')
        programs = guard.state.programs_by_name(method)
        yield (
            no_programs_for_rpc(method, args)
            if programs.empty else
            run_programs_exclusive(guard, programs, args)
        )
    return handler


__all__ = ('rpc_handler', 'run_program', 'run_programs', 'run_program_exclusive', 'run_programs_exclusive',
           'no_programs_for_rpc', 'decode_args', 'rpc_handler',)



from typing import TypeVar, Any

from amino import Nil
from amino.do import do, Do

from ribosome.config.config import Config
from ribosome.logging import nvim_logging
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.config.component import ComponentConfig
from ribosome.components.internal.update import init_rpc_plugin
from ribosome.compute.interpret import ProgIOInterpreter
from ribosome.compute.program import Program

D = TypeVar('D')
CC = TypeVar('CC')


def cons_state(
        config: Config,
        io_interpreter: ProgIOInterpreter=None,
        logger: Program[None]=None,
        **kw: Any,
) -> PluginState[D, CC]:
    data = config.basic.state_ctor()
    return PluginState.cons(
        config.basic,
        ComponentConfig(config.components),
        config.rpc,
        data,
        Nil,
        config.init,
        logger=logger,
        io_interpreter=io_interpreter,
        **kw,
    )


@do(NvimIO[PluginState[D, CC]])
def init_state_plugin(config: Config, io_interpreter: ProgIOInterpreter=None, logger: Program[None]=None) -> Do:
    log_handler = yield N.delay(nvim_logging)
    state = cons_state(config, io_interpreter, logger, log_handler=log_handler)
    yield init_rpc_plugin().run_s(state)


__all__ = ('cons_state', 'init_state',)



from typing import Generic, TypeVar, Callable
from threading import Event
from queue import Queue

from amino import ADT, Dat, List, Nil, Maybe

from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog

A = TypeVar('A')
B = TypeVar('B')


class InputChar(ADT['InputChar']):
    pass


class PrintableChar(InputChar):

    def __init__(self, char: str) -> None:
        self.char = char


class SpecialChar(InputChar):

    def __init__(self, char: str, modifiers: List[str]) -> None:
        self.char = char
        self.modifiers = modifiers


class Input(ADT['Input']):
    pass


class NoInput(Input):
    pass


class NormalInput(Input):

    def __init__(self, char: InputChar) -> None:
        self.char = char


class InterruptInput(Input):
    pass


class PromptInputAction(Generic[A], ADT['PromptInputAction[A]']):
    pass


class PromptInit(PromptInputAction[A]):
    pass


class PromptInput(PromptInputAction[A]):

    def __init__(self, char: InputChar) -> None:
        self.char = char


class PromptInterrupt(PromptInputAction[A]):
    pass


class PromptConsumerInput(PromptInputAction[A]):

    def __init__(self, data: A) -> None:
        self.data = data


class PromptUpdate(Generic[A], ADT['PromptUpdate[A]']):
    pass


class PromptUpdateInit(PromptUpdate[A]):
    pass


class PromptUpdateChar(PromptUpdate[A]):

    def __init__(self, char: InputChar) -> None:
        self.char = char


class PromptUpdateConsumer(PromptUpdate[A]):

    def __init__(self, data: A) -> None:
        self.data = data


class Prompt(Dat['Prompt']):

    @staticmethod
    def cons(cursor: int=0, pre: str='', post: str='') -> 'Prompt':
        return Prompt(cursor, pre, post)

    def __init__(self, cursor: int, pre: str, post: str) -> None:
        self.cursor = cursor
        self.pre = pre
        self.post = post

    @property
    def line(self) -> str:
        return self.pre + self.post


class PromptState(ADT['PromptState']):
    pass


class PromptEcho(PromptState):
    pass


class PromptPassthrough(PromptState):
    pass


class PromptAction(ADT['PromptAction']):
    pass


class PromptStateTrans(PromptAction):

    def __init__(self, state: PromptState) -> None:
        self.state = state


class PromptQuit(PromptAction):
    pass


class PromptUnit(PromptAction):
    pass


class InputState(Generic[A, B], Dat['InputState[A, B]']):

    @staticmethod
    def cons(
            data: A,
            actions: List[PromptUpdate[B]]=Nil,
            prompt: Prompt=Prompt.cons(),
            cursor: int=0,
            state: PromptState=PromptEcho(),
    ) -> 'InputState':
        return InputState(
            data,
            actions,
            prompt,
            cursor,
            state,
        )

    def __init__(
            self,
            data: A,
            actions: List[PromptUpdate[B]],
            prompt: Prompt,
            cursor: int,
            state: PromptState,
    ) -> None:
        self.data = data
        self.actions = actions
        self.prompt = prompt
        self.cursor = cursor
        self.state = state


ProcessPrompt = Callable[[PromptUpdate[B]], NS[InputState[A, B], PromptAction]]


class InputResources(Generic[A, B], Dat['InputResources[A, B]']):

    @staticmethod
    def cons(
            state: InputState[A, B],
            update: Callable[[PromptUpdate[B]], NS[InputState[A, B], None]],
            inputs: Queue=None,
            stop: Event=None,
            interval: float=.01,
    ) -> 'InputResources[A, B]':
        return InputResources(
            state,
            update,
            inputs or Queue(),
            stop or Event(),
            interval,
        )

    def __init__(
            self,
            state: InputState[A, B],
            update: Callable[[PromptUpdate[B]], NS[InputState[A, B], None]],
            inputs: Queue,
            stop: Event,
            interval: float,
    ) -> None:
        self.state = state
        self.update = update
        self.inputs = inputs
        self.stop = stop
        self.interval = interval


__all__ = ('InputChar', 'PrintableChar', 'SpecialChar', 'Input', 'NoInput', 'NormalInput', 'InterruptInput',
           'InputState', 'InputResources', 'InputChar', 'PromptInputAction', 'PromptInput', 'PromptInterrupt',
           'Prompt', 'PromptState', 'PromptEcho', 'PromptQuit', 'PromptPassthrough', 'PromptUnit', 'PromptAction',
           'PromptStateTrans', 'PromptUnit', 'PromptQuitWith', 'PromptConsumerInput', 'PromptUpdate',
           'PromptUpdateChar', 'PromptUpdateConsumer', 'PromptInit', 'PromptUpdateInit',)



from typing import Any, Tuple

from amino import do, Do, IO, List, Dat, Map, Path
from amino.test import temp_dir
from amino.test.path import pkg_dir
from amino.env_vars import set_env

from ribosome import NvimApi
from ribosome.rpc.uv.uv import cons_uv_embed, cons_uv_socket
from ribosome.rpc.start import start_external
from ribosome.rpc.comm import Comm
from ribosome.test.config import TestConfig
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.variable import variable_set
from ribosome.rpc.nvim_api import RiboNvimApi

nvim_cmdline = List('nvim', '-n', '-u', 'NONE')
env_vars = List(
    ('RIBOSOME_SPEC', 1),
    ('AMINO_DEVELOPMENT', 1),
    ('RIBOSOME_PKG_DIR', str(pkg_dir())),
    ('RIBOSOME_FILE_LOG_FMT', '{levelname} {name}:{message}')
)


def test_env_vars(config: TestConfig) -> List[Tuple[str, Any]]:
    return env_vars.cat(('RIBOSOME_LOG_FILE', str(config.log_file)))


class TestNvimApi(RiboNvimApi):

    def __init__(self, name: str, comm: Comm, config: TestConfig) -> None:
        self.name = name
        self.comm = comm
        self.config = config


class TestNvim(Dat['TestNvim']):

    def __init__(self, comm: Comm, api: NvimApi) -> None:
        self.comm = comm
        self.api = api


@do(IO[None])
def setup_env(config: TestConfig) -> Do:
    yield test_env_vars(config).traverse2(set_env, IO)
    yield IO.delay(config.log_dir.mkdir, exist_ok=True, parents=True)
    yield IO.delay(config.log_file.touch)
    yield set_env('RIBOSOME_LOG_FILE', str(config.log_file))


def start_uv_embed(config: TestConfig) -> IO[NvimApi]:
    ''' start an embedded vim session that loads no init.vim.
    `_temp/log/vim` is set as log file. aside from being convenient, this is crucially necessary, as the first use of
    the session will block if stdout is used for output.
    '''
    log = temp_dir('log') / 'vim'
    argv = nvim_cmdline + List('--embed', f'-V{log}')
    uv, rpc_comm = cons_uv_embed(argv)
    return start_external(config.config.basic.name, rpc_comm)


def start_uv_socket(config: TestConfig, socket: Path) -> IO[NvimApi]:
    uv, rpc_comm = cons_uv_socket(socket)
    return start_external(config.config.basic.name, rpc_comm)


@do(IO[TestNvim])
def setup_test_nvim_embed(config: TestConfig) -> Do:
    yield setup_env(config)
    api = yield start_uv_embed(config)
    return TestNvim(api.comm, TestNvimApi(api.name, api.comm, config))


@do(IO[TestNvim])
def setup_test_nvim_socket(config: TestConfig, socket: Path) -> Do:
    yield setup_env(config)
    api = yield start_uv_socket(config, socket)
    return TestNvim(api.comm, TestNvimApi(api.name, api.comm, config))


def set_nvim_vars(vars: Map[str, Any]) -> NvimIO[None]:
    return vars.to_list.traverse2(variable_set, NvimIO)


__all__ = ('TestNvim', 'setup_test_nvim_embed', 'set_nvim_vars',)



from amino import do, Do, List
from amino.json import dump_json

from ribosome.nvim.io.compute import NvimIO
from ribosome.test.config import TestConfig
from ribosome.nvim.api.function import define_function, nvim_call_function
from ribosome.config.config import Config
from ribosome.nvim.io.api import N

stderr_handler_name = 'RibosomeSpecStderr'
stderr_handler_body = '''
let err = substitute(join(a:data, '\\r'), '"', '\\"', 'g')
try
    python3 import amino
    python3 from ribosome.logging import ribosome_envvar_file_logging
    python3 ribosome_envvar_file_logging()
    execute 'python3 amino.amino_log.error(f"""error starting rpc job on channel ' . a:id . ':\\r' . err . '""")'
catch //
    echoerr err
endtry
'''


def start_plugin_cmd_import(path: str) -> NvimIO[str]:
    return N.pure(f'from ribosome.host import start_path; start_path({path!r})')


@do(NvimIO[str])
def start_plugin_cmd_json(config: Config) -> Do:
    json = yield N.from_either(dump_json(config))
    return f'from ribosome.host import start_json_config; start_json_config({json!r})'


@do(NvimIO[None])
def start_plugin_embed(config: TestConfig) -> Do:
    yield define_function(stderr_handler_name, List('id', 'data', 'event'), stderr_handler_body)
    cmd = yield (
        config.config_path.cata(
            start_plugin_cmd_import,
            lambda: start_plugin_cmd_json(config.config),
        )
    )
    args = ['python3', '-c', cmd]
    opts = dict(rpc=True, on_stderr=stderr_handler_name)
    yield nvim_call_function('jobstart', args, opts)


__all__ = ('start_plugin_embed',)



from amino.options import env_xdg_data_dir
from amino import Either, Path, Right, Try, Nil, Left, Eval, Do
from amino.do import do
from amino.boolean import true

from ribosome.nvim.io.compute import NvimIO
from ribosome.config.setting import EvalSetting, list_setting, path_setting, str_setting, bool_setting
from ribosome.nvim.io.api import N
from ribosome.nvim.api.rpc import plugin_name


@do(NvimIO[Either[str, Path]])
def state_dir_with_name() -> Do:
    plugin = yield plugin_name()
    base = yield state_dir.value_or_default()
    pro_name = yield proteome_main_name.value
    sess_name = yield ribosome_session_name.value
    path = pro_name.o(sess_name) / (lambda a: base / plugin / a)
    yield N.pure(path)


state_dir_help = '''This directory is used to persist the plugin's current state.'''
state_dir_base_default = env_xdg_data_dir.value / Path | (Path.home() / '.local' / 'share')
proteome_name_help = 'If **proteome** is installed, the session name is obtained from the main project name.'
session_name_help = 'A custom session name for the state dir can be specified.'
components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `@prog`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom components.
Users can use this variable to inject programs into any plugin that can use the plugin's data.
'''

components = list_setting('components', 'names or paths of active components', components_help, True, Right(Nil))
state_dir = path_setting('ribosome_state_dir', 'state persistence directory', state_dir_help, False,
                         Right(state_dir_base_default))
proteome_main_name = str_setting('proteome_main_name', 'project name from protoeome', proteome_name_help, False)
ribosome_session_name = str_setting('ribosome_session_name', 'project name from user var', session_name_help, False)
project_state_dir = EvalSetting('project_state_dir', Eval.always(state_dir_with_name))
run_internal_init = bool_setting('run_internal_init', 'run internal initialization handler', '', True, Right(true))


__all__ = ('components', 'state_dir', 'proteome_main_name', 'ribosome_session_name', 'project_state_dir',
           'run_internal_init')



from typing import Callable, TypeVar, Tuple

from amino import IO, do, Do, Nil
from amino.state import State
from amino.logging import module_log

from ribosome.rpc.comm import Comm, Rpc, RpcComm, StateGuard
from ribosome.config.config import Config
from ribosome.rpc.handle_receive import rpc_receive
from ribosome.components.internal.update import init_rpc_plugin
from ribosome.rpc.state import cons_state
from ribosome.nvim.io.compute import NvimIO, NvimIOSuspend
from ribosome.nvim.io.api import N
from ribosome.nvim.api.variable import variable_set_prefixed
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PS
from ribosome.logging import nvim_logging
from ribosome.rpc.to_vim import rpc_error
from ribosome.rpc.to_plugin import rpc_handler
from ribosome.rpc.from_vim import execute_rpc_from_vim
from ribosome.rpc.nvim_api import RiboNvimApi
from ribosome.components.internal.prog import internal_init
from ribosome.compute.run import run_prog

A = TypeVar('A')
log = module_log()


def start_comm(comm: Comm, execute_request: Callable[[Comm, Rpc], None]) -> IO[None]:
    return comm.rpc.start_processing(rpc_receive(comm, execute_request), rpc_error(comm))


def stop_comm(comm: Comm) -> IO[None]:
    return IO.delay(comm.rpc.stop_processing)


@do(NS[PS, None])
def init_plugin() -> Do:
    yield NS.delay(nvim_logging)
    yield init_rpc_plugin()
    yield run_prog(internal_init, Nil)
    yield NS.lift(variable_set_prefixed('started', True))


@do(IO[None])
def init_comm(rpc_comm: RpcComm, execute_request: Callable[[Comm, Rpc], IO[None]]) -> Do:
    comm = Comm.cons(rpc_handler, rpc_comm)
    yield start_comm(comm, execute_request)
    return comm


def plugin_execute_receive_request(guard: StateGuard[A], plugin_name: str) -> Callable[[Comm, Rpc], IO[None]]:
    def execute(comm: Comm, rpc: Rpc) -> IO[None]:
        return IO.fork_io(execute_rpc_from_vim, rpc, comm, comm.request_handler(guard), plugin_name, daemon=True)
    return execute


@do(NvimIO[Tuple[Comm, StateGuard]])
def setup_comm(config: Config, rpc_comm: RpcComm) -> Do:
    state = cons_state(config)
    guard = StateGuard.cons(state)
    execute_request = plugin_execute_receive_request(guard, config.basic.name)
    comm = yield N.from_io(init_comm(rpc_comm, execute_request))
    api = RiboNvimApi(config.basic.name, comm)
    yield NvimIOSuspend.cons(State.set(api).replace(N.pure(None)))
    return comm, guard


@do(NvimIO[None])
def start_plugin(config: Config, rpc_comm: RpcComm) -> Do:
    comm, guard = yield setup_comm(config, rpc_comm)
    state = yield init_plugin().run_s(guard.state)
    yield N.delay(lambda v: guard.init(state))
    return comm


@do(IO[None])
def start_plugin_sync(config: Config, rpc_comm: RpcComm) -> Do:
    start = start_plugin(config, rpc_comm)
    result = yield IO.delay(start.run_a, None)
    comm = yield IO.from_either(result.to_either)
    yield comm.rpc.join()


def cannot_execute_request(comm: Comm, rpc: Rpc) -> IO[None]:
    return IO.failed(f'cannot execute request in external nvim: {rpc}')


@do(IO[RiboNvimApi])
def start_external(name: str, rpc_comm: RpcComm) -> Do:
    comm = yield init_comm(rpc_comm, cannot_execute_request)
    return RiboNvimApi(name, comm)


__all__ = ('start_comm', 'stop_comm', 'init_plugin', 'init_comm', 'plugin_execute_receive_request', 'setup_comm',
           'start_plugin', 'start_plugin_sync', 'cannot_execute_request', 'start_external',)





def escape_squote(text: str) -> str:
    return text.replace("'", "''")


def escape_dquote(text: str) -> str:
    return text.replace('"', '\\"')


def escape_quote(text: str) -> str:
    return escape_dquote(escape_squote(text))


def quote_for_ex(text: str) -> str:
    return f'\'{escape_quote(text)}\''


__all__ = ('escape_squote', 'escape_dquote', 'escape_quote',)



from amino import ADT


class RpcType(ADT['RpcType']):
    pass


class BlockingRpc(RpcType):

    def __init__(self, id: int) -> None:
        self.id = id


class NonblockingRpc(RpcType):
    pass


__all__ = ('RpcType', 'BlockingRpc', 'NonblockingRpc',)



from amino.boolean import true
from amino import List, Just

from ribosome.components.internal.prog import (program_log, set_log_level, update_state, update_component_state,
                                               state_data, rpc_triggers, poll, append_python_path, show_python_path,
                                               enable_components, mapping, internal_init, component_state_data,
                                               rpc_job_stderr)
from ribosome.config.component import Component
from ribosome.rpc.data.prefix_style import Full
from ribosome.rpc.api import rpc

internal: Component = Component.cons(
    'internal',
    rpc=List(
        rpc.read(program_log).conf(prefix=Full(), sync=true),
        rpc.write(set_log_level).conf(prefix=Full()),
        rpc.write(update_state).conf(json=true),
        rpc.write(update_component_state).conf(json=true),
        rpc.read(state_data).conf(name=Just('state'), prefix=Full()),
        rpc.write(component_state_data).conf(name=Just('component_state'), prefix=Full()),
        rpc.read(rpc_triggers).conf(internal=true, sync=true, prefix=Full()),
        rpc.read(poll).conf(prefix=Full()),
        rpc.write(append_python_path).conf(prefix=Full()),
        rpc.read(show_python_path).conf(prefix=Full()),
        rpc.write(enable_components).conf(prefix=Full()),
        rpc.write(mapping).conf(name=Just('map'), prefix=Full()),
        rpc.write(internal_init).conf(prefix=Full()),
        rpc.read(rpc_job_stderr).conf(prefix=Full()),
    ),
)

__all__ = ('internal',)



import abc
from pathlib import Path
import shutil
from contextlib import contextmanager
from typing import Tuple, Awaitable, Any, Generic, TypeVar
from subprocess import PIPE, Popen

import asyncio
from asyncio.subprocess import PIPE as APIPE

from amino import Map, Future, __, Boolean, _, L, List, Maybe, IO, Lists, do, Either
from amino.lazy import lazy
from amino.either import Right, Left
from amino.util.string import ToStr
from amino.dat import Dat
from amino.do import Do
from amino.logging import module_log

import ribosome
from ribosome.logging import Logging
from ribosome.nvim.api.data import NvimApi

log = module_log()


class Result(ToStr):

    def __init__(self, job: 'Job', success: bool, out: str, err: str) -> None:
        self.job = job
        self.success = Boolean(success)
        self.out = out
        self.err = err

    def _arg_desc(self) -> List[str]:
        return List(str(self.job), str(self.success), self.msg)

    @property
    def msg(self):
        return self.err if self.err else self.out

    def either(self, good, bad):
        return self.success.maybe(Right(good)) | Left(bad)


class JobClient(Dat['JobClient']):

    def __init__(self, cwd: Path, name: str) -> None:
        self.cwd = cwd
        self.name = name


class Job(Dat['Job']):

    def __init__(self, client: JobClient, exe: str, args: List, kw: Map, loop: Any, pipe_in: Maybe[str]) -> None:
        self.client = client
        self.exe = exe
        self.args = args
        self.kw = kw
        self.loop = loop
        self.pipe_in = pipe_in

    @lazy
    def status(self):
        return Future(loop=self.loop)

    @lazy
    def stdin(self):
        return self.pipe_in.map(__.encode()) | None

    def finish(self, f: Future) -> None:
        code, out, err = f.result()
        self.finish_strict(code, out, err)

    def finish_strict(self, code: int, out: str, err: str) -> None:
        self.status.set_result(self.result_strict(code, out, err))

    def result_strict(self, code: int, out: str, err: str) -> None:
        return Result(self, code == 0, out, err)

    def finish_success(self, out: str) -> None:
        self.finish_strict(0, out, '')

    def cancel(self, reason: str):
        self.status.set_result(Result(self, False, '', 'canceled: {}'.format(reason)))

    @property
    def valid(self):
        return (
            not self.status.done() and
            self.cwd.is_dir() and (
                Path(self.exe).exists or
                shutil.which(self.exe) is not None
            )
        )

    @property
    def cwd(self):
        return self.client.cwd

    def __str__(self):
        return 'Job({}, {}, {})'.format(self.client.name, self.exe, ' '.join(self.args))

    def run(self):
        self.loop.run_until_complete(self.status)

    @property
    def result(self):
        return self.status.result()

    @property
    def success(self):
        return self.status.done() and self.result.success


@contextmanager
def _dummy_ctx():
    yield


class ProcessExecutor(Logging, abc.ABC):
    ''' Handler for subprocess execution
    Because python handles signals only on the main thread and
    subprocess notifcations, like their termination, is noticed by
    catching SIG_CHLD, the actual execution of the coroutines spawning
    the subprocess must be done while the main threads's event loop is
    in the running state, so it can relay the signal to the subprocess's
    event loop.
    As this class can be instantiated and used from within another
    thread's event loop, it is also impossible (afaict) to use either
    that loop or the main thread loop, for proc.wait() thus blocks
    indefinitely with a <defunct> process.
    '''

    def __init__(self, loop=None, main_loop_ctx=_dummy_ctx) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.main_loop_ctx = main_loop_ctx
        self.current: Map[Any, Job] = Map()

    def process(self, job: Job) -> Awaitable:
        self.log.debug(f'creating subprocess for {job}')
        return asyncio.create_subprocess_exec(
            job.exe,
            *job.args,
            stdin=APIPE,
            stdout=APIPE,
            stderr=APIPE,
            cwd=str(job.cwd),
            loop=self.loop,
            **job.kw,
        )

    async def _execute(self, job: Job):
        def error(msg: str) -> Tuple[int, str, str]:
            return -111, '', msg
        try:
            out, err = None, None
            with self.main_loop_ctx():
                proc = await self.process(job)
                self.log.debug(f'awaiting {job} on the main loop')
                out, err = await proc.communicate(job.stdin)
            if out is None or err is None:
                return error('executing {} failed'.format(job))
            else:
                self.log.debug(f'{job} executed successfully ({out}, {err})')
                return proc.returncode, out.decode(), err.decode()
        except Exception as e:
            self.log.exception(f'{job} failed with {e!r}')
            return error(f'exception: {e}')

    def run(self, job: Job) -> Future[Result]:
        ''' return values of execute are set as result of the task
        returned by ensure_future(), obtainable via task.result()
        '''
        if not self.watcher_ready:
            self.log.error(f'child watcher unattached when executing {job}')
            job.cancel('unattached watcher')
        elif not self.can_execute(job):
            self.log.error('invalid execution job: {}'.format(job))
            job.cancel('invalid')
        else:
            self.log.debug('executing {}'.format(job))
            task = asyncio.ensure_future(self._execute(job), loop=self.loop)
            task.add_done_callback(job.finish)
            task.add_done_callback(L(self.job_done)(job, _))
            self.current[job.client] = job
        return job.status

    @property
    def watcher_ready(self) -> bool:
        with self.main_loop_ctx():
            watcher = asyncio.get_child_watcher()
        return (
            watcher is not None and
            watcher._loop is not None
        )

    def can_execute(self, job: Job):
        return (
            job.client not in self.current and
            job.valid
        )

    def job_done(self, job, status):
        self.log.debug('{} is done with status {}'.format(job, status))
        if job.client in self.current:
            self.current.pop(job.client)

    @property
    def ready(self):
        return self.current.is_empty

    def _run_jobs(self):
        self.current.valmap(lambda job: job.run())

    @property
    def futures(self):
        return self.current.v.map(_.status)

    def await_threadsafe(self, loop):
        async def waiter():
            while not self.ready:
                await asyncio.sleep(0.001)
        loop.run_until_complete(waiter())


class NvimProcessExecutor(ProcessExecutor):

    def __init__(self, vim: NvimApi, loop=None) -> None:
        self.vim = vim
        super().__init__(loop, self._main_event_loop)

    def _main_event_loop(self):
        return self.vim.threadsafe_subprocess() if ribosome.in_vim else _dummy_ctx()


A = TypeVar('A')


class SubprocessResult(Generic[A], Dat['SubprocessResult']):

    def __init__(self, retval: int, stdout: List[str], stderr: List[str], data: A) -> None:
        self.retval = retval
        self.stdout = stdout
        self.stderr = stderr
        self.data = data

    @property
    def success(self) -> Boolean:
        return Boolean(self.retval == 0)


class Subprocess(Generic[A], Dat['Subprocess']):

    @staticmethod
    @do(IO[Tuple[int, List[str], List[str]]])
    def popen(*args: str, timeout: float=None, stdin: int=PIPE, stdout: int=PIPE, stderr: int=PIPE, env: dict=dict(),
              universal_newlines=True, **kw: Any) -> Do:
        log.debug(f'executing subprocess `{args}`')
        pop = yield IO.delay(Popen, args, stdin=stdin, stdout=stdout, stderr=stderr, env=env,
                             universal_newlines=universal_newlines, **kw)
        out, err = yield IO.delay(pop.communicate, timeout=timeout)
        out_lines = Lists.lines(out or '')
        err_lines = Lists.lines(err or '')
        yield IO.pure((-1 if pop.returncode is None else pop.returncode, out_lines, err_lines))

    def __init__(self, exe: Path, args: List[str], data: A, timeout: float, **kw: Any) -> None:
        self.exe = exe
        self.args = args
        self.data = data
        self.timeout = timeout
        self.kw = kw

    @property
    def args_tuple(self) -> tuple:
        return tuple(self.args.cons(str(self.exe)))

    @do(IO[SubprocessResult[A]])
    def execute(self, **kw: Any) -> Do:
        retval, out, err = yield Subprocess.popen(*self.args_tuple, timeout=self.timeout, **kw, **self.kw)
        yield IO.pure(SubprocessResult(retval, out, err, self.data))


__all__ = ('ProcessExecutor', 'Result', 'Job', 'NvimProcessExecutor', 'Subprocess')



from typing import Generic, TypeVar, Callable, Any

from amino import Dat, ADT, List, Nil, Maybe
from amino.logging import module_log

from ribosome.util.menu.prompt.data import PromptState, InputState, PromptUpdate
from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog
from ribosome.nvim.scratch import ScratchBuffer

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')


class MenuLine(Generic[A], Dat['MenuLine[A]']):

    @staticmethod
    def cons(
            text: str,
            meta: A,
            visible: bool=True,
            selected: bool=False,
    ) -> 'MenuLine[A]':
        return MenuLine(
            text,
            meta,
            visible,
            selected,
        )

    def __init__(self, text: str, meta: A, visible: bool, selected: bool) -> None:
        self.text = text
        self.meta = meta
        self.visible = visible
        self.selected = selected


class MenuContent(Generic[A], Dat['MenuContent[A]']):

    @staticmethod
    def cons(
            lines: List[MenuLine[A]]=Nil,
            visible: List[int]=Nil,
    ) -> 'MenuContent[A]':
        return MenuContent(lines, visible)

    def __init__(self, lines: List[MenuLine[A]], visible: List[int]) -> None:
        self.lines = lines
        self.visible = visible


def visible_lines(content: MenuContent[A]) -> List[MenuLine[A]]:
    return content.lines.filter(lambda a: a.visible)


def selected_lines(content: MenuContent[A], cursor: int) -> List[MenuLine[A]]:
    visible = content.lines.filter(lambda a: a.visible)
    selected = visible.filter(lambda a: a.selected)
    return (
        visible.lift(cursor).to_list
        if selected.empty else
        selected
    )


class MenuAction(ADT['MenuAction']):
    pass


class MenuQuit(MenuAction):
    pass


class MenuQuitWith(MenuAction):

    def __init__(self, next: Prog[None]) -> None:
        self.next = next


class MenuPrompt(MenuAction):

    def __init__(self, state: PromptState) -> None:
        self.state = state


class MenuUpdateLines(MenuAction):

    def __init__(self, content: MenuContent) -> None:
        self.content = content


class MenuUpdateCursor(MenuAction):
    pass


class MenuPush(MenuAction):

    def __init__(self, thunk: Callable[[ScratchBuffer], NS[Any, Prog[None]]]) -> None:
        self.thunk = thunk


class MenuPop(MenuAction):
    pass


class MenuUnit(MenuAction):
    pass


class MenuStackAction(ADT['MenuStackAction']):
    pass


class MenuStackQuit(MenuStackAction):
    pass


class MenuStackPush(MenuStackAction):

    def __init__(self, current: NS[Any, Prog[None]], thunk: NS[Any, Prog[None]]) -> None:
        self.current = current
        self.thunk = thunk


class MenuStackPop(MenuStackAction):
    pass


class MenuState(Generic[S, ML], Dat['MenuState[S, ML]']):

    @staticmethod
    def cons(
            state: S,
            content: MenuContent[ML]=None,
            cursor: int=0,
            next: MenuStackAction=MenuStackQuit(),
            result: Prog[None]=None,
    ) -> 'MenuState[S, ML]':
        return MenuState(
            state,
            content or MenuContent.cons(),
            cursor,
            next,
            Maybe.optional(result),
        )

    def __init__(
            self,
            state: S,
            content: MenuContent[ML],
            cursor: int,
            next: MenuStackAction,
            result: Maybe[Prog[None]],
    ) -> None:
        self.state = state
        self.content = content
        self.cursor = cursor
        self.next = next
        self.result = result


class MenuConfig(Generic[S, ML, U], Dat['MenuConfig[S, ML, U]']):

    @staticmethod
    def cons(
            handle_input: Callable[[PromptUpdate[U]], NS[InputState[MenuState[S, ML], U], MenuAction]],
            name: str='',
    ) -> 'MenuConfig[S, ML, U]':
        return MenuConfig(
            handle_input,
            name,
        )

    def __init__(
            self,
            handle_input: Callable[[PromptUpdate[U]], NS[InputState[MenuState[S, ML], U], MenuAction]],
            name: str,
    ) -> None:
        self.handle_input = handle_input
        self.name = name


class Menu(Generic[S, ML, U], Dat['Menu[S, ML, U]']):

    @staticmethod
    def cons(
            config: MenuConfig[S, ML, U],
            state: MenuState[S, ML],
    ) -> 'Menu':
        return Menu(
            config,
            state,
        )

    def __init__(self, config: MenuConfig[S, ML, U], state: MenuState[S, ML]) -> None:
        self.config = config
        self.state = state


class Menus(Dat['Menus']):

    @staticmethod
    def cons(
            stack: List[NS[Any, Prog[None]]]=Nil,
    ) -> 'Menus':
        return Menus(stack)

    def __init__(self, stack: List[NS[Any, Prog[None]]]) -> None:
        self.stack = stack


__all__ = ('MenuConfig', 'MenuLine', 'MenuState', 'Menu', 'MenuContent', 'MenuAction', 'MenuQuit', 'MenuPrompt',
           'MenuUpdateLines', 'MenuUnit', 'MenuUpdateCursor', 'visible_lines', 'selected_lines', 'MenuPush', 'MenuPop',
           'Menus', 'MenuStackAction', 'MenuStackQuit', 'MenuStackPush', 'MenuStackPop',)



from typing import Callable, Any

from kallikrein import Expectation

from amino import Do, do
from amino.lenses.lens import lens

from ribosome.nvim.io.compute import NvimIO
from ribosome.test.config import TestConfig
from ribosome.nvim.api.data import StrictNvimApi
from ribosome.test.run import run_test_io
from ribosome.nvim.io.api import N
from ribosome.test.prog import init_test_state
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PS
from ribosome.test.request import StrictRequestHandler


def setup_strict_test_nvim(config: TestConfig) -> StrictNvimApi:
    handler = StrictRequestHandler(config.request_handler, config.function_handler, config.command_handler)
    return StrictNvimApi.cons(config.config.basic.name, vars=config.vars, request_handler=handler)


def unit_test(config: TestConfig, io: Callable[..., NS[PS, Expectation]], *a: Any, **kw: Any) -> Expectation:
    initial_nvim = setup_strict_test_nvim(config)
    @do(NvimIO[Expectation])
    def run() -> Do:
        state = yield init_test_state(config)
        yield io(*a, **kw).run_a(state)
    return run_test_io(N.to_io_a, run(), initial_nvim)


def update_data(**kw: Any) -> NS[PS, None]:
    return NS.modify(lens.data.modify(lambda a: a.copy(**kw)))


__all__ = ('unit_test', 'update_data',)



from typing import TypeVar, Callable

from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances, tc_prop
from amino.lazy import lazy
from amino import Map
from amino.tc.monoid import Monoid

from ribosome.nvim.io.compute import NvimIO, flat_map_nvim_io, NvimIOPure

A = TypeVar('A')
B = TypeVar('B')


class NvimIOInstances(ImplicitInstances):

    @lazy
    def _instances(self) -> Map:
        return Map({
            Monad: NvimIOMonad(),
            Monoid: NvimIOMonoid(),
        })


class NvimIOMonad(Monad):

    def pure(self, a: A) -> NvimIO[A]:
        return NvimIOPure(a)

    def flat_map(self, fa: NvimIO[A], f: Callable[[A], NvimIO[B]]) -> NvimIO[B]:
        return flat_map_nvim_io(f)(fa)


class NvimIOMonoid(Monoid):

    @tc_prop
    def empty(self) -> NvimIO[None]:
        return NvimIOPure(None)

    def combine(self, a: NvimIO[A], b: NvimIO[A]) -> NvimIO[A]:
        return a.flat_map(b)


__all__ = ('NvimIOInstances', 'NvimIOMonad', 'NvimIOMonoid',)



from typing import Any, Callable

from kallikrein import Expectation, k
from kallikrein.matchers.either import be_right
from kallikrein.matchers.match_with import match_with

from amino import IO

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.variable import pvar_becomes


def run_test_io(io: Callable[..., IO[Expectation]], *a: Any, **kw: Any) -> Expectation:
    return k(io(*a, **kw).attempt).must(be_right(match_with(lambda a: a)))


def plugin_started() -> NvimIO[None]:
    return pvar_becomes('started', True, timeout=5)


__all__ = ('run_test_io', 'plugin_started',)



from typing import TypeVar, Callable

from amino import do, Do, Map, Nil, List, Just, Nothing
from amino.case import Case
from amino.lenses.lens import lens
from amino.logging import module_log

from ribosome.util.menu.data import (Menu, MenuConfig, MenuState, MenuContent, MenuUpdateLines, MenuLine,
                                     selected_lines, MenuUnit)
from ribosome.util.menu.prompt.data import (InputChar, InputState, PromptUpdate, PromptUpdateChar, PromptUpdateConsumer,
                                            PrintableChar, SpecialChar, PromptUpdateInit, PromptEcho)
from ribosome.nvim.io.state import NS
from ribosome.util.menu.auto.data import AutoUpdate, AutoUpdateRefresh, AutoUpdateConsumer, AutoState, AutoS
from ribosome.util.menu.auto.cmd import builtin_mappings

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')


@do(NS[InputState[A, B], C])
def prompt_state_fork(
        echo: Callable[[], NS[InputState[A, B], C]],
        passthrough: Callable[[], NS[InputState[A, B], C]],
) -> Do:
    state = yield NS.inspect(lambda a: a.state)
    yield (
        echo()
        if isinstance(state, PromptEcho) else
        passthrough()
    )


class auto_update(Case[AutoUpdate[U, ML], AutoS], alg=AutoUpdate):

    @do(AutoS)
    def refresh(self, update: AutoUpdateRefresh[U, ML]) -> Do:
        yield NS.unit

    @do(AutoS)
    def consumer(self, update: AutoUpdateConsumer[U, ML]) -> Do:
        consumer = yield NS.inspect(lambda a: a.data.state.consumer)
        yield consumer(PromptUpdateConsumer(update))


@do(AutoS)
def filter_menu(char: str) -> Do:
    lines = yield NS.inspect(lambda a: a.data.content.lines)
    filter = yield NS.inspect(lambda a: a.prompt.line)
    updated_lines = lines.map(lambda a: a.set.visible(filter in a.text))
    visible = lines.with_index.collect(lambda a: Just(a[0]) if a[1].visible else Nothing)
    yield NS.modify(lens.data.content.modify(lambda a: a.copy(lines=updated_lines, visible=visible)))
    content = yield NS.inspect(lambda a: a.data.content)
    return MenuUpdateLines(content)


@do(AutoS)
def menu_mapping(code: str, char: InputChar) -> Do:
    mappings = yield NS.inspect(lambda a: a.data.state.mappings)
    consumer = yield NS.inspect(lambda a: a.data.state.consumer)
    yield mappings.lift(code).map(lambda a: a()).get_or(consumer, PromptUpdateChar(char))


class auto_menu_input_char(Case[InputChar, AutoS[U, ML, S]], alg=InputChar):

    def printable(self, char: PrintableChar) -> AutoS:
        return prompt_state_fork(lambda: filter_menu(char.char), lambda: menu_mapping(char.char, char))

    def special(self, char: SpecialChar) -> AutoS:
        return menu_mapping(char.char, char)


class auto_menu_handle(Case[PromptUpdate[AutoUpdate[U, ML]], AutoS[U, ML, S]], alg=PromptUpdate):

    @do(AutoS)
    def init(self, update: PromptUpdateInit[AutoUpdate[U, ML]]) -> Do:
        content = yield NS.inspect(lambda a: a.data.content)
        yield NS.pure(MenuUpdateLines(content))

    @do(AutoS)
    def char(self, update: PromptUpdateChar[AutoUpdate[U, ML]]) -> Do:
        yield auto_menu_input_char.match(update.char)

    @do(AutoS)
    def consumer(self, update: PromptUpdateConsumer[AutoUpdate[U, ML]]) -> Do:
        yield auto_update.match(update.data)


def no_process(update: PromptUpdate[AutoUpdate[U, ML]]) -> AutoS:
    return NS.pure(MenuUnit())


def auto_menu(
        state: S,
        content: MenuContent[ML],
        name: str,
        mappings: Map[str, Callable[[], AutoS]]=Map(),
        consumer: Callable[[PromptUpdate[AutoUpdate[U, ML]]], AutoS]=no_process,
) -> Menu[AutoState[U, ML, S], ML, C]:
    config: MenuConfig[AutoState[U, ML, S], ML, C] = MenuConfig.cons(auto_menu_handle.match, name)
    all_mappings = builtin_mappings ** mappings
    menu_state: MenuState[AutoState[U, ML, S], ML] = MenuState.cons(AutoState(consumer, state, all_mappings), content)
    return Menu.cons(config, menu_state)


@do(NS[InputState[MenuState[AutoState[U, ML, S], ML], AutoUpdate[U, ML]], List[MenuLine[ML]]])
def selected_menu_lines() -> Do:
    content = yield NS.inspect(lambda a: a.data.content)
    cursor = yield NS.inspect(lambda a: a.data.cursor)
    return selected_lines(content, cursor)


__all__ = ('auto_menu', 'selected_menu_lines',)



from setuptools import setup, find_packages

version_parts = (13, 0, 0, 'a', 76)
version = '.'.join(map(str, version_parts))

setup(
    name='ribosome',
    description='neovim plugin framework',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/ribosome',
    packages=find_packages(exclude=['unit', 'unit.*', 'integration', 'integration.*', 'test', 'test.*']),
    install_requires=[
        'amino~=13.0.1a7',
        'pyuv~=1.4.0',
        'msgpack-python~=0.5.6'
    ],
    tests_require=[
        'chiasma~=0.1.0.a28',
        'kallikrein~=0.22.0a15',
    ],
    entry_points={
        'console_scripts': [
            'ribosome_start_plugin = ribosome.cli:start_plugin',
        ],
    },
)



from typing import Any, TypeVar, Callable, Generic

from amino import List, Dat, Maybe, Map

from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.program import Program
from ribosome.rpc.data.prefix_style import PrefixStyle, Short, Plain
from ribosome.rpc.data.rpc_method import RpcMethod, CommandMethod, FunctionMethod, AutocmdMethod
from ribosome.util.doc.data import DocBlock

A = TypeVar('A')
default_methods = List(FunctionMethod(), CommandMethod.cons())


class RpcOptions(Dat['RpcOptions']):

    @staticmethod
    def cons(
            name: str=None,
            methods: List[RpcMethod]=default_methods,
            prefix: PrefixStyle=Short(),
            json: bool=False,
            write: bool=True,
            help: DocBlock=None,
            params_help: List[str]=None,
            json_help: Map[str, str]=None,
    ) -> 'RpcOptions':
        return RpcOptions(
            Maybe.optional(name),
            methods,
            prefix,
            json,
            write,
            help or DocBlock.empty(),
            Maybe.optional(params_help),
            Maybe.optional(json_help),
        )


    def __init__(
            self,
            name: Maybe[str],
            methods: List[RpcMethod],
            prefix: PrefixStyle,
            json: bool,
            write: bool,
            help: DocBlock,
            params_help: Maybe[List[str]],
            json_help: Maybe[Map[str, str]],
    ) -> None:
        self.name = name
        self.methods = methods
        self.prefix = prefix
        self.json = json
        self.write = write
        self.help = help
        self.params_help = params_help
        self.json_help = json_help


class RpcProgram(Generic[A], Dat['RpcProgram[A]']):

    @staticmethod
    def cons(program: Program[A], options: RpcOptions=None) -> 'RpcProgram':
        return RpcProgram(program, options or RpcOptions.cons())

    def __init__(self, program: Program[A], options: RpcOptions) -> None:
        self.program = program
        self.options = options

    def conf(self, **kw: Any) -> 'RpcProgram':
        return self.set.options(self.options.copy(**kw))

    @property
    def program_name(self) -> str:
        return self.program.name

    @property
    def rpc_name(self) -> str:
        return self.options.name | self.program_name


class RpcApi:

    def simple(self, f: Callable[..., NvimIO[A]]) -> RpcProgram[A]:
        ...

    def write(self, program: Program[A]) -> RpcProgram[A]:
        return RpcProgram.cons(program)

    def read(self, program: Program[A]) -> RpcProgram[A]:
        return RpcProgram.cons(program, RpcOptions.cons(write=False))

    def autocmd(self, program: Program[A], pattern: str=None, sync: bool=False) -> RpcProgram[A]:
        method = AutocmdMethod.cons(pattern, sync)
        return RpcProgram.cons(program, RpcOptions.cons(methods=List(method))).conf(prefix=Plain())


rpc = RpcApi()


__all__ = ('RpcOptions', 'RpcProgram', 'rpc',)
